#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

script_variable = Struct("script_variable",
    ascii_str32("script_variable_name"),
    ascii_str32("exposed_parameter_name"),
    SIZE=80
    )

effect = Struct("effect",
    dependency_os("effect", "efpg"),
    ascii_str32("name"),
    reflexive("script_variables", script_variable, 32,
        DYN_NAME_PATH='.script_variable_name'),
    SIZE=72
    )

activation_control = Struct("activation_control",
    SEnum16("state",
        "initially_active",
        "is_in_cutscene",
        "player_is_zoomed",
        "player_is_using_a_vehicle"
        ),
    Bool16("flags",
        "invert"
        ),
    SIZE=4
    )

effect_instance = Struct("effect_instance",
    ascii_str32("name"),
    dyn_senum16("effect",
        DYN_NAME_PATH="tagdata.effects.STEPTREE[DYN_I].effect.filepath"),
    SEnum16("render_stage",
        "after_bsp_and_before_blur",
        "after_blur_and_before_alphad_faces",
        "after_alphad_faces_and_before_hud",
        "after_hud_and_before_menu",
        "after_menu",
        ),
    QStruct("quad_tesselation",
        SInt16("x"), SInt16("y"), ORIENT="h"
        ),
    from_to_zero_to_one("x_screen_bounds"),
    from_to_zero_to_one("y_screen_bounds"),

    Pad(12),
    SEnum16("activation_operation",
        "all",
        "any"
        ),

    Pad(6),
    reflexive("activation_controls", activation_control, 8),

    SIZE=116
    )

efpc_body = Struct("tagdata",
    Pad(12),
    # since there APPEARS to be a reflexive in the tag, but it
    # doesnt show up in Guerilla, I'm gonna turn it into padding
    # and leave this as a reminder that a reflexive might exist.
    #reflexive("unknown", void_desc, 0),
    Pad(12),
    reflexive("effects", effect, 32,
        DYN_NAME_PATH='.name'),
    reflexive("effect_instances", effect_instance, 32,
        DYN_NAME_PATH='.name'),
    SIZE=48
    )

def get():
    return efpc_def

efpc_def = TagDef("efpc",
    blam_header_os('efpc'),
    efpc_body,

    ext=".effect_postprocess_collection", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import pi, sqrt, log
from struct import Struct as PyStruct
from traceback import format_exc

from reclaimer.meta.wrappers.halo_map import HaloMap
from reclaimer.meta.wrappers.halo1_rsrc_map import Halo1RsrcMap
from reclaimer.meta.wrappers.halo1_map import Halo1Map
from reclaimer import data_extraction
from reclaimer.halo_script.hsc import h1_script_syntax_data_def
from reclaimer.hek.defs.coll import fast_coll_def
from reclaimer.hek.defs.sbsp import fast_sbsp_def
from reclaimer.hek.handler import HaloHandler
from reclaimer.util import int_to_fourcc

from supyr_struct.field_types import FieldType

__all__ = ("Halo1AnniMap",)


def end_swap_float(v, packer=PyStruct(">f").pack,
                   unpacker=PyStruct("<f").unpack):
    return unpacker(packer(v))[0]


def end_swap_int32(v):
    assert v >= -0x80000000 and v < 0x80000000
    if v < 0:
        v += 0x100000000
    v = ((((v << 24) + (v >> 24)) & 0xFF0000FF) +
         ((v << 8) & 0xFF0000) +
         ((v >> 8) & 0xFF00))
    if v & 0x80000000:
        return v - 0x100000000
    return v


def end_swap_int16(v):
    assert v >= -0x8000 and v < 0x8000
    if v < 0:
        v += 0x10000
    v = ((v << 8)  + (v >> 8)) & 0xFFFF
    if v & 0x8000:
        return v - 0x10000
    return v


def end_swap_uint32(v):
    assert v >= 0 and v <= 0xFFFFFFFF
    return ((((v << 24) + (v >> 24)) & 0xFF0000FF) +
            ((v << 8) & 0xFF0000) +
            ((v >> 8) & 0xFF00))


def end_swap_uint16(v):
    assert v >= 0 and v <= 0xFFFF
    return ((v << 8) + (v >> 8)) & 0xFFFF


class Halo1AnniMap(Halo1Map):
    tag_headers = None
    defs = None

    handler_class = HaloHandler

    inject_rawdata = Halo1RsrcMap.inject_rawdata

    def __init__(self, maps=None):
        HaloMap.__init__(self, maps)
        self.setup_tag_headers()

    def get_dependencies(self, meta, tag_id, tag_cls):
        if self.is_indexed(tag_id):
            if tag_cls != "snd!":
                return ()

            rsrc_id = meta.promotion_sound.id & 0xFFff
            if rsrc_id == 0xFFFF: return ()

            sounds = self.maps.get("sounds")
            rsrc_id = rsrc_id // 2
            if   sounds is None: return ()
            elif rsrc_id >= len(sounds.tag_index.tag_index): return ()

            tag_path = sounds.tag_index.tag_index[rsrc_id].path
            inv_snd_map = getattr(self, 'ce_tag_indexs_by_paths', {})
            tag_id = inv_snd_map.get(tag_path, 0xFFFF)
            if tag_id >= len(self.tag_index.tag_index): return ()

            return [self.tag_index.tag_index[tag_id]]

        if self.handler is None: return ()

        dependency_cache = self.handler.tag_ref_cache.get(tag_cls)
        if not dependency_cache: return ()

        nodes = self.handler.get_nodes_by_paths(dependency_cache, (None, meta))
        dependencies = []

        for node in nodes:
            if node.id & 0xFFff == 0xFFFF:
                continue
            dependencies.append(node)
        return dependencies

    def get_meta(self, tag_id, reextract=False, ignore_rsrc_sounds=False, **kw):
        '''
        Takes a tag reference id as the sole argument.
        Returns that tags meta data as a parsed block.
        '''
        if tag_id is None:
            return
        magic     = self.map_magic
        map_data  = self.map_data
        tag_index = self.tag_index
        tag_index_array = tag_index.tag_index

        tag_id &= 0xFFFF
        tag_index_ref = self.tag_index_manager.get_tag_index_ref(tag_id)
        if tag_index_ref is None:
            return

        tag_cls = None
        if tag_id == tag_index.scenario_tag_id & 0xFFff:
            tag_cls = "scnr"
        elif tag_index_ref.class_1.enum_name not in ("<INVALID>", "NONE"):
            tag_cls = int_to_fourcc(tag_index_ref.class_1.data)

        # if we dont have a defintion for this tag_cls, then return nothing
        if self.get_meta_descriptor(tag_cls) is None:
            return


        if tag_cls is None:
            # couldn't determine the tag class
            return
        elif reextract:
            pass
        elif tag_id == tag_index.scenario_tag_id & 0xFFff and self.scnr_meta:
            return self.scnr_meta
        elif tag_cls == "matg" and self.matg_meta:
            return self.matg_meta

        desc = self.get_meta_descriptor(tag_cls)
        block = [None]
        pointer_converter = self.bsp_pointer_converters.get(
            tag_id, self.map_pointer_converter)

        offset = pointer_converter.v_ptr_to_f_ptr(tag_index_ref.meta_offset)

        try:
            # read the meta data from the map
            with FieldType.force_big:
                desc['TYPE'].parser(
                    desc, parent=block, attr_index=0, rawdata=map_data,
                    map_pointer_converter=pointer_converter,
                    tag_index_manager=self.tag_index_manager, offset=offset)
        except Exception:
            print(format_exc())
            if kw.get("allow_corrupt"):
                return block[0]
            return

        meta = block[0]
        try:

            self.record_map_cache_read(tag_id, 0)  # cant get size quickly enough
            if self.map_cache_over_limit():
                self.clear_map_cache()

            self.inject_rawdata(meta, tag_cls, tag_index_ref)
            self.byteswap_anniversary_fields(meta, tag_cls)
        except Exception:
            print(format_exc())
            if not kw.get("allow_corrupt"):
                meta = None

        return meta

    def byteswap_anniversary_fields(self, meta, tag_cls):

        if tag_cls == "antr":
            unpack_header = PyStruct("<11i").unpack
            for b in meta.animations.STEPTREE:
                b.unknown_sint16 = end_swap_int16(b.unknown_sint16)
                b.unknown_float = end_swap_float(b.unknown_float)
                if not b.flags.compressed_data:
                    continue

                comp_data = b.frame_data.data[b.offset_to_compressed_data: ]
                # byteswap compressed frame data header
                for i in range(0, 44, 4):
                    data = comp_data[i: i + 4]
                    for j in range(4):
                        comp_data[i + 3 - j] = data[j]

                header = list(unpack_header(comp_data[: 44]))
                header.insert(0, 44)
                header.append(len(comp_data))

                item_sizes = (4, 2, 2, 2,  4, 2, 4, 4,  4, 2, 4, 4)
                comp_data_off_len_size = []
                for i in range(len(header) - 1):
                    comp_data_off_len_size.append([
                        header[i], header[i + 1] - header[i], item_sizes[i]])

                for off, length, size in comp_data_off_len_size:
                    for i in range(off, off + length, size):
                        data = comp_data[i: i + size]
                        for j in range(size):
                            comp_data[i + size - 1 - j] = data[j]

                # replace the frame_data with the compressed data and some
                # blank default / frame data so tool doesnt shit the bed.
                default_data_size = b.node_count * (12 + 8 + 4) - b.frame_size
                b.default_data.data += bytearray(
                    max(0, len(b.default_data.data) - default_data_size))

                b.offset_to_compressed_data = b.frame_count * b.frame_size
                b.frame_data.data = bytearray(
                    b.frame_count * b.frame_size) + comp_data

        elif tag_cls == "bitm":
            for b in meta.bitmaps.STEPTREE:
                b.pixels = end_swap_uint16(b.pixels)

        elif tag_cls == "coll":
            for b in meta.nodes.STEPTREE:
                b.unknown0 = end_swap_int16(b.unknown0)
                b.unknown1 = end_swap_int16(b.unknown1)

        elif tag_cls == "effe":
            for event in meta.events.STEPTREE:
                for b in event.particles.STEPTREE:
                    b.unknown0 = end_swap_int16(b.unknown0)
                    b.unknown1 = end_swap_int16(b.unknown1)

        elif tag_cls == "hmt ":
            block_bytes = bytearray(meta.string.serialize())
            for i in range(20, len(block_bytes), 2):
                byte = block_bytes[i + 1]
                block_bytes[i + 1] = block_bytes[i]
                block_bytes[i] = byte

            meta.string.parse(rawdata=block_bytes)

        elif tag_cls == "lens":
            meta.unknown0 = end_swap_float(meta.unknown0)
            meta.unknown1 = end_swap_float(meta.unknown1)

        elif tag_cls == "lsnd":
            meta.unknown0 = end_swap_float(meta.unknown0)
            meta.unknown1 = end_swap_float(meta.unknown1)
            meta.unknown2 = end_swap_float(meta.unknown2)
            meta.unknown3 = end_swap_float(meta.unknown3)
            meta.unknown4 = end_swap_int16(meta.unknown4)
            meta.unknown5 = end_swap_int16(meta.unknown5)
            meta.unknown6 = end_swap_float(meta.unknown6)

        elif tag_cls == "metr":
            meta.screen_x_pos = end_swap_uint16(meta.screen_x_pos)
            meta.screen_y_pos = end_swap_uint16(meta.screen_y_pos)
            meta.width  = end_swap_uint16(meta.width)
            meta.height = end_swap_uint16(meta.height)

        elif tag_cls in ("mod2", "mode"):
            for node in meta.nodes.STEPTREE:
                node.unknown = end_swap_float(node.unknown)
                for b in (node.rot_jj_kk, node.rot_kk_ii, node.rot_ii_jj,
                          node.translation_to_root):
                    for i in range(len(b)):
                        b[i] = end_swap_float(b[i])

        elif tag_cls == "part":
            meta.rendering.unknown0 = end_swap_int32(meta.rendering.unknown0)
            meta.rendering.unknown1 = end_swap_float(meta.rendering.unknown1)
            meta.rendering.unknown2 = end_swap_uint32(meta.rendering.unknown2)

        elif tag_cls == "pphy":
            meta.scaled_density = end_swap_float(meta.scaled_density)
            meta.water_gravity_scale = end_swap_float(meta.water_gravity_scale)
            meta.air_gravity_scale = end_swap_float(meta.air_gravity_scale)

        elif tag_cls == "sbsp":
            # TODO: Might need to byteswap cluster data and sound_pas data

            for coll_mat in meta.collision_materials.STEPTREE:
                coll_mat.unknown = end_swap_uint32(coll_mat.unknown)

            node_data = meta.nodes.STEPTREE
            for i in range(0, len(node_data), 2):
                b0 = node_data[i]
                node_data[i] = node_data[i + 1]
                node_data[i + 1] = b0

            leaf_data = meta.leaves.STEPTREE
            for i in range(0, len(leaf_data), 16):
                b0 = leaf_data[i]
                leaf_data[i] = leaf_data[i + 1]
                leaf_data[i + 1] = b0

                b0 = leaf_data[i + 2]
                leaf_data[i + 2] = leaf_data[i + 3]
                leaf_data[i + 3] = b0

                b0 = leaf_data[i + 4]
                leaf_data[i + 4] = leaf_data[i + 5]
                leaf_data[i + 5] = b0

                b0 = leaf_data[i + 6]
                leaf_data[i + 6] = leaf_data[i + 7]
                leaf_data[i + 7] = b0

            for lightmap in meta.lightmaps.STEPTREE:
                for b in lightmap.materials.STEPTREE:
                    vt_ct = b.vertices_count
                    l_vt_ct = b.lightmap_vertices_count

                    u_verts = b.uncompressed_vertices.STEPTREE
                    c_verts = b.compressed_vertices.STEPTREE

                    b.unknown_meta_offset0 = end_swap_uint32(
                        b.unknown_meta_offset0)
                    b.vertices_meta_offset = end_swap_uint32(
                        b.vertices_meta_offset)

                    b.vertex_type.data = end_swap_uint16(b.vertex_type.data)

                    b.unknown_meta_offset1 = end_swap_uint32(
                        b.unknown_meta_offset1)
                    b.lightmap_vertices_meta_offset = end_swap_uint32(
                        b.lightmap_vertices_meta_offset)

                    # byteswap (un)compressed verts and lightmap verts
                    for data in (u_verts, c_verts):
                        for i in range(0, len(data), 4):
                            b0 = data[i]
                            b1 = data[i+1]
                            data[i]   = data[i+3]
                            data[i+1] = data[i+2]
                            data[i+2] = b1
                            data[i+3] = b0

                    # since the compressed lightmap u and v coordinates are
                    # 2 byte fields rather than 4, the above byteswapping
                    # will have swapped u and v. we need to swap them back.
                    # multiply vt_ct by 32 to skip non-lightmap verts, and
                    # add 4 to skip the 4 byte compressed lightmap normal.
                    for i in range(vt_ct * 32 + 4, len(c_verts), 8):
                        c_verts[i: i + 1] = c_verts[i+1], c_verts[i]

            for fog_plane in meta.fog_planes.STEPTREE:
                fog_plane.material_type.data = end_swap_int16(
                    fog_plane.material_type.data)

        elif tag_cls == "scnr":
            for b in meta.object_names.STEPTREE:
                b.object_type.data = end_swap_uint16(b.object_type.data)
                b.reflexive_index = end_swap_int16(b.reflexive_index)

            for b in meta.trigger_volumes.STEPTREE:
                b.unknown = end_swap_uint16(b.unknown)

            for b in meta.encounters.STEPTREE:
                b.unknown = end_swap_int16(b.unknown)

            # PROLLY GONNA HAVE TO BYTESWAP RECORDED ANIMS AND MORE SHIT
            syntax_data = meta.script_syntax_data.data
            with FieldType.force_big:
                syntax_header = h1_script_syntax_data_def.build(rawdata=syntax_data)

            i = 56
            for node_i in range(syntax_header.last_node):
                n_typ = syntax_data[i + 5] + (syntax_data[i + 4] << 8)
                flags = syntax_data[i + 7] + (syntax_data[i + 6] << 8)
                if flags & 7 == 1:
                    # node is a primitive
                    if n_typ == 5:
                        # node is a boolean
                        syntax_data[i + 19] = syntax_data[i + 16]
                        syntax_data[i + 16: i + 19] = (0, 0, 0) # null these 3
                    elif n_typ == 7:
                        # node is a sint16
                        syntax_data[i + 18] = syntax_data[i + 16]
                        syntax_data[i + 19] = syntax_data[i + 17]
                        syntax_data[i + 16: i + 18] = (0, 0) # null these 2

                i += 20

        elif tag_cls == "senv":
            meta.senv_attrs.bump_properties.map_scale_x = end_swap_float(
                meta.senv_attrs.bump_properties.map_scale_x)
            meta.senv_attrs.bump_properties.map_scale_y = end_swap_float(
                meta.senv_attrs.bump_properties.map_scale_y)

        elif tag_cls == "snd!":
            for pr in meta.pitch_ranges.STEPTREE:
                for b in pr.permutations.STEPTREE:
                    b.ogg_sample_count = end_swap_uint32(b.ogg_sample_count)

        elif tag_cls == "spla":
            meta.spla_attrs.primary_noise_map.unknown0 = end_swap_uint16(
                meta.spla_attrs.primary_noise_map.unknown0)
            meta.spla_attrs.primary_noise_map.unknown1 = end_swap_uint16(
                meta.spla_attrs.primary_noise_map.unknown1)

            meta.spla_attrs.secondary_noise_map.unknown0 = end_swap_uint16(
                meta.spla_attrs.secondary_noise_map.unknown0)
            meta.spla_attrs.secondary_noise_map.unknown1 = end_swap_uint16(
                meta.spla_attrs.secondary_noise_map.unknown1)

        elif tag_cls == "ustr":
            for b in meta.strings.STEPTREE:
                block_bytes = bytearray(b.serialize())
                for i in range(12, len(block_bytes), 2):
                    byte = block_bytes[i + 1]
                    block_bytes[i + 1] = block_bytes[i]
                    block_bytes[i] = byte

                b.parse(rawdata=block_bytes)


        if tag_cls in ("bipd", "vehi", "weap", "eqip", "garb", "proj",
                       "scen", "mach", "ctrl", "lifi", "plac", "ssce", "obje"):
            meta.obje_attrs.object_type.data = end_swap_int16(
                meta.obje_attrs.object_type.data)
        elif tag_cls in ("senv", "soso", "sotr", "schi", "scex",
                         "swat", "sgla", "smet", "spla", "shdr"):
            meta.shdr_attrs.shader_type.data = end_swap_int16(
                meta.shdr_attrs.shader_type.data)

    def inject_rawdata(self, meta, tag_cls, tag_index_ref):
        pass

    def meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs):
        kwargs["byteswap"] = False
        Halo1Map.meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs)


        # TODO: Remove this if bitmap/sound extraction is ever implemented
        if tag_cls in ("snd!", "bitm"):
            # these tags don't properly extract due to missing
            # pixel data and sound permutation sample data
            return
        elif tag_cls in ("sbsp", "scnr"):
            # renderable geometry is absent from h1 anniversary bsps, and
            # we don't know how to properly byteswap recorded animations,
            # so scenarios can't be extracted properly either.
            return

        return meta



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.bipd import *
from .obje import *
from .unit import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=0)

bipd_body = Struct("tagdata",
    obje_attrs,
    unit_attrs,
    bipd_attrs,
    SIZE=1268,
    )

def get():
    return bipd_def

bipd_def = TagDef("bipd",
    blam_header('bipd', 3),
    bipd_body,

    ext=".biped", endian=">", tag_cls=BipdTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.common_descs import *
from supyr_struct.defs.tag_def import TagDef

hsc__body = Struct("tagdata",
    ascii_str32("name"),
    h2_rawtext_ref("source"),
    ENDIAN="<", SIZE=40
    )


def get():
    return hsc__def

hsc__def = TagDef("hsc*",
    h2_blam_header('hsc*'),
    hsc__body,

    ext=".%s" % h2_tag_class_fcc_to_ext["hsc*"], endian="<"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.meta.halo1_map_fast_functions import shader_class_bytes,\
     object_class_bytes, read_reflexive, read_rawdata_ref, iter_reflexive_offs,\
     repair_dependency, repair_dependency_array
from reclaimer.halo_script.hsc import get_hsc_data_block,\
     HSC_IS_SCRIPT_OR_GLOBAL
#from supyr_struct.util import *

MAX_MATERIAL_COUNT = 33


def get_tagc_refs(meta_offset, map_data, magic, tag_classes_by_id, tag_index_array):
    try:
        ct, moff, _ = read_reflexive(map_data, meta_offset - magic, 200, 16, magic)
    except Exception:
        return [], []

    if ct > 200 or ct <= 0 or moff != meta_offset + 12:
        return [], []

    # might be Soul, but need to check tag classes to make sure
    try:
        reffed_ids = []
        reffed_types = []
        for moff2 in iter_reflexive_offs(
                map_data, meta_offset - magic, 16, 200, magic):
            map_data.seek(moff2 - magic + 12)
            tag_id = int.from_bytes(map_data.read(4), "little")
            if tag_id == 0xFFffFFff:
                continue
            elif ((tag_id & 0xFFff) not in range(len(tag_index_array)) or
                  tag_index_array[tag_id & 0xFFff].id != tag_id):
                # break on the first invalid tag id.
                # the reflexive size might be corrupt
                break

            reffed_ids.append(tag_id & 0xFFff)
            reffed_types.append(tag_classes_by_id.get(tag_id & 0xFFff))

        return reffed_ids, reffed_types
    except Exception:
        pass

    return [], []


def repair_hud_background(index_array, map_data, magic, repair, engine, offset):
    args = (index_array, map_data, magic, repair, engine, b'mtib')
    repair_dependency(*(args + (offset + 36, )))
    # multitex overlays
    for moff in iter_reflexive_offs(map_data, offset + 88 - magic, 480, 30, magic):
        repair_dependency(*(args + (moff + 100, )))
        repair_dependency(*(args + (moff + 116, )))
        repair_dependency(*(args + (moff + 132, )))


def repair_devi_attrs(offset, index_array, map_data, magic, repair, engine):
    # struct size is 276
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (None, offset + 36)))
    repair_dependency(*(args + (None, offset + 52)))
    repair_dependency(*(args + (None, offset + 68)))
    repair_dependency(*(args + (None, offset + 84)))
    repair_dependency(*(args + (None, offset + 100)))
    repair_dependency(*(args + (None, offset + 116)))
    repair_dependency(*(args + (None, offset + 144)))


def repair_item_attrs(offset, index_array, map_data, magic, repair, engine):
    # struct size is 396
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'toof', offset + 204)))
    repair_dependency(*(args + (b'!dns', offset + 220)))
    repair_dependency(*(args + (b'effe', offset + 364)))
    repair_dependency(*(args + (b'effe', offset + 380)))


def repair_unit_attrs(offset, index_array, map_data, magic, repair, engine):
    # struct size is 372
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'effe', offset + 12)))

    # camera tracks
    ct, moff, _ = read_reflexive(map_data, offset + 120 - magic, 2, 28, magic)
    repair_dependency_array(*(args + (b'kart', moff, ct, 28)))

    repair_dependency(*(args + (b'vtca', offset + 208)))
    repair_dependency(*(args + (b'!tpj', offset + 268)))

    # new hud interfaces
    ct, moff, _ = read_reflexive(map_data, offset + 300 - magic, 2, 48, magic)
    repair_dependency_array(*(args + (b'ihnu', moff, ct, 48)))

    # dialogue variants
    ct, moff, _ = read_reflexive(map_data, offset + 312 - magic, 16, 24, magic)
    repair_dependency_array(*(args + (b'gldu', moff + 8, ct, 24)))

    # weapons
    ct, moff, _ = read_reflexive(map_data, offset + 348 - magic, 4, 36, magic)
    repair_dependency_array(*(args + (b'paew', moff, ct, 36)))

    # seats
    for moff in iter_reflexive_offs(map_data, offset + 360 - magic, 284, 16, magic):
        # camera tracks
        ct, moff2, _ = read_reflexive(map_data, moff + 208 - magic, 2, 28, magic)
        repair_dependency_array(*(args + (b'kart', moff2, ct, 28)))

        # new hud interfaces
        ct, moff2, _ = read_reflexive(map_data, moff + 220 - magic, 2, 48, magic)
        repair_dependency_array(*(args + (b'ihnu', moff2, ct, 48)))

        repair_dependency(*(args + (b'vtca', moff + 248)))

        if "yelo" in engine:
            # seat extension
            for moff2 in iter_reflexive_offs(map_data, moff + 264 - magic, 100, 1, magic):
                # seat boarding
                for moff3 in iter_reflexive_offs(map_data, moff2 + 28 - magic, 112, 1, magic):
                    # seat keyframe action
                    for moff4 in iter_reflexive_offs(map_data, moff3 + 76 - magic, 152, 12, magic):
                        repair_dependency(*(args + (b'!tpj', moff4 + 48)))
                        repair_dependency(*(args + (b'effe', moff4 + 68)))

                # seat damage
                for moff3 in iter_reflexive_offs(map_data, moff2 + 40 - magic, 136, 1, magic):
                    repair_dependency(*(args + (b'!tpj', moff3 + 4)))
                    repair_dependency(*(args + (b'!tpj', moff3 + 96)))

    if "yelo" in engine:
        # unit extension
        for moff in iter_reflexive_offs(map_data, offset + 288 - magic, 60, 1, magic):
            # mounted states
            for moff2 in iter_reflexive_offs(map_data, moff - magic, 128, 1, magic):
                # camera tracks
                ct, moff3, _ = read_reflexive(map_data, moff2 + 80 - magic, 2, 28, magic)
                repair_dependency_array(*(args + (b'kart', moff3, ct, 28)))

                # unit keyframe action
                for moff3 in iter_reflexive_offs(map_data, moff2 + 92, 96, 12, magic):
                    repair_dependency(*(args + (b'!tpj', moff3 + 8)))
                    repair_dependency(*(args + (b'effe', moff3 + 24)))


def repair_actv(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'rtca', tag_offset + 0x4)))
    repair_dependency(*(args + (b'tinu', tag_offset + 0x14)))
    repair_dependency(*(args + (b'vtca', tag_offset + 0x24)))
    repair_dependency(*(args + (b'paew', tag_offset + 0x64)))
    repair_dependency(*(args + (b'piqe', tag_offset + 0x1C0)))


def repair_ant_(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'mtib', tag_offset + 0x20)))
    repair_dependency(*(args + (b'yhpp', tag_offset + 0x30)))


def repair_antr(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    ct, moff, _ = read_reflexive(
        map_data, index_array[tag_id].meta_offset + 0x54 - magic, 257, 20, magic)
    repair_dependency_array(index_array, map_data, magic, repair, engine,
                            b'!dns', moff, ct, 20)


def repair_coll(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine, b'effe')
    repair_dependency(*(args + (tag_offset + 0x70, )))
    repair_dependency(*(args + (tag_offset + 0x84, )))
    repair_dependency(*(args + (tag_offset + 0x98, )))
    repair_dependency(*(args + (tag_offset + 0xA8, )))
    repair_dependency(*(args + (tag_offset + 0xBC, )))
    repair_dependency(*(args + (tag_offset + 0x188, )))
    repair_dependency(*(args + (tag_offset + 0x198, )))
    repair_dependency(*(args + (tag_offset + 0x1A8, )))

    # regions
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0x240 - magic, 8, 0x54, magic)
    repair_dependency_array(*(args + (moff + 0x38, ct, 0x54)))


def repair_cont(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'mtib', tag_offset + 0x30)))
    repair_dependency(*(args + (b'mtib', tag_offset + 0xD0)))

    # point states
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0x138 - magic, 16, 104, magic)
    repair_dependency_array(*(args + (b'yhpp', moff + 16, ct, 104)))


def repair_deca(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    moff = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'aced', moff + 0x8)))
    repair_dependency(*(args + (b'mtib', moff + 0xD8)))


def repair_DeLa(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'mtib', tag_offset + 56)))

    # event handlers
    for moff in iter_reflexive_offs(map_data, tag_offset + 84 - magic, 72, 32, magic):
        repair_dependency(*(args + (b'aLeD', moff + 8)))
        repair_dependency(*(args + (b'!dns', moff + 24)))

    # search and replace functions
    # Not needed anymore when tags are parsed in safe-mode
    #ct, _, __ = read_reflexive(map_data, tag_offset + 96 - magic)
    #if ct > 32:
    #    # some people apparently think its cute to set this reflexive
    #    # count so high so that tool just fails to compile the tag
    #    map_data.seek(tag_offset + 96 - magic)
    #    map_data.write(b'\x20\x00\x00\x00')

    repair_dependency(*(args + (b'rtsu', tag_offset + 236)))
    repair_dependency(*(args + (b'tnof', tag_offset + 252)))
    repair_dependency(*(args + (b'mtib', tag_offset + 340)))
    repair_dependency(*(args + (b'mtib', tag_offset + 356)))
    repair_dependency(*(args + (b'aLeD', tag_offset + 420)))

    # conditional widgets
    ct, moff, _ = read_reflexive(map_data, tag_offset + 724 - magic, 32, 80, magic)
    repair_dependency_array(*(args + (b'aLeD', moff, ct, 80)))

    # child widgets
    ct, moff, _ = read_reflexive(map_data, tag_offset + 992 - magic, 32, 80, magic)
    repair_dependency_array(*(args + (b'aLeD', moff, ct, 80)))


def repair_dobc(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'mtib', index_array[tag_id].meta_offset + 52)


def repair_effe(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    # events
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x34 - magic, 0x44, 32, magic):
        # parts
        for moff2 in iter_reflexive_offs(map_data, moff + 0x2C - magic, 0x68, 32, magic):
            map_data.seek(moff2 - magic + 0x14)
            tag_class = map_data.read(4)
            repair_dependency(*(args + (tag_class, moff2 + 0x18)))

        # particles
        ct, moff2, _ = read_reflexive(map_data, moff + 0x38 - magic, 32, 0xE8, magic)
        repair_dependency_array(*(args + (b'trap', moff2 + 0x54, ct, 0xE8)))


def repair_elec(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'mtib', index_array[tag_id].meta_offset + 52)

def repair_flag(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'rdhs', tag_offset + 0x18)))
    repair_dependency(*(args + (b'yhpp', tag_offset + 0x28)))
    repair_dependency(*(args + (b'rdhs', tag_offset + 0x44)))


def repair_fog(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'mtib', tag_offset + 0xBC)))
    repair_dependency(*(args + (b'dnsl', tag_offset + 0xF4)))
    repair_dependency(*(args + (b'edns', tag_offset + 0x104)))


def repair_font(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency_array(index_array, map_data, magic, repair, engine,
                            b'mtib', index_array[tag_id].meta_offset + 0x3C, 4)


def repair_foot(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    # effects
    for moff in iter_reflexive_offs(map_data, tag_offset - magic, 28, 13, magic):
        # materials
        for moff2 in iter_reflexive_offs(map_data, moff - magic, 48, MAX_MATERIAL_COUNT, magic):
            repair_dependency(*(args + (b'effe', moff2)))
            repair_dependency(*(args + (b'!dns', moff2 + 16)))


def repair_glw_(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'mtib', index_array[tag_id].meta_offset + 324)

def repair_grhi(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # grenade hud background
    repair_hud_background(*(args + (tag_offset + 36, )))

    # total grenades background
    repair_hud_background(*(args + (tag_offset + 140, )))

    repair_dependency(*(args + (b'mtib', tag_offset + 332)))

    # warning sounds
    for moff in iter_reflexive_offs(map_data, tag_offset + 360 - magic, 56, 12, magic):
        map_data.seek(moff - magic)
        if map_data.read(4) == b'dnsl':
            repair_dependency(*(args + (b'dnsl', moff)))
        else:
            repair_dependency(*(args + (b'!dns', moff)))


def repair_hud_(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'mtib', index_array[tag_id].meta_offset)


def repair_hudg(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'tnof', tag_offset + 0x48, )))
    repair_dependency(*(args + (b'tnof', tag_offset + 0x58, )))
    repair_dependency(*(args + (b'rtsu', tag_offset + 0x94, )))
    repair_dependency(*(args + (b'mtib', tag_offset + 0xA4, )))
    repair_dependency(*(args + (b'rtsu', tag_offset + 0xB4, )))
    repair_dependency(*(args + (b' tmh', tag_offset + 0xF0, )))
    repair_dependency(*(args + (b'mtib', tag_offset + 0x150, )))
    repair_dependency(*(args + (b'ihpw', tag_offset + 0x2C0, )))
    repair_dependency(*(args + (b'mtib', tag_offset + 0x338, )))
    repair_dependency(*(args + (b'mtib', tag_offset + 0x3C8, )))
    repair_dependency(*(args + (b'!dns', tag_offset + 0x3E0, )))


def repair_itmc(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # item permutations
    ct, moff, _ = read_reflexive(map_data, tag_offset - magic, 32767, 84, magic)
    repair_dependency_array(*(args + (b'meti', moff + 36, ct, 84)))


def repair_jpt_(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'!dns', index_array[tag_id].meta_offset + 0x114)


def repair_lens(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'mtib', index_array[tag_id].meta_offset + 0x20)


def repair_ligh(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'mtib', tag_offset + 0x64)))
    repair_dependency(*(args + (b'mtib', tag_offset + 0x7C)))
    repair_dependency(*(args + (b'snel', tag_offset + 0xAC)))


def repair_lsnd(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'gmdc', tag_offset + 0x2C)))

    # tracks
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x3C - magic, 0xA0, 4, magic):
        repair_dependency(*(args + (b'!dns', moff + 0x30)))
        repair_dependency(*(args + (b'!dns', moff + 0x40)))
        repair_dependency(*(args + (b'!dns', moff + 0x50)))
        repair_dependency(*(args + (b'!dns', moff + 0x80)))
        repair_dependency(*(args + (b'!dns', moff + 0x90)))

    # detail sounds
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0x48 - magic, 32, 104, magic)
    repair_dependency_array(*(args + (b'!dns', moff, ct, 104)))


def repair_matg(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # sounds
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0xF8 - magic, 2, 16, magic)
    repair_dependency_array(*(args + (b'!dns', moff, ct)))

    # camera
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0x104 - magic, 1, 16, magic)
    repair_dependency_array(*(args + (b'kart', moff, ct)))

    # grenades
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x128 - magic, 68, 4, magic):
        repair_dependency(*(args + (b'effe', moff + 4)))
        repair_dependency(*(args + (b'ihrg', moff + 20)))
        repair_dependency(*(args + (b'piqe', moff + 36)))
        repair_dependency(*(args + (b'jorp', moff + 52)))

    # rasterizer data
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x134 - magic, 428, 1, magic):
        # function textures
        repair_dependency_array(*(args + (b'mtib', moff, 7)))
        moff += 7*16 + 60
        # default/experimental/video effect textures
        repair_dependency_array(*(args + (b'mtib', moff, 9)))
        moff += 9*16 + 52 + 4*11
        # pc textures
        repair_dependency(*(args + (b'mtib', moff)))

    # interface bitmaps
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x140 - magic, 304, 1, magic):
        repair_dependency_array(*(args + (b'tnof', moff, 2)))
        repair_dependency_array(*(args + (b'oloc', moff + 32, 4)))
        repair_dependency(*(args + (b'gduh', moff + 96)))
        repair_dependency_array(*(args + (b'mtib', moff + 112, 3)))
        repair_dependency(*(args + (b'#rts', moff + 160)))
        repair_dependency(*(args + (b'#duh', moff + 176)))
        repair_dependency_array(*(args + (b'mtib', moff + 192, 4)))

    # weapons
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0x14C - magic, 20, 16, magic)
    repair_dependency_array(*(args + (b'ejbo', moff, ct)))

    # powerups
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0x158 - magic, 20, 16, magic)
    repair_dependency_array(*(args + (b'ejbo', moff, ct)))

    # multiplayer info
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x164 - magic, 160, 1, magic):
        repair_dependency(*(args + (b'meti', moff)))
        repair_dependency(*(args + (b'tinu', moff + 16)))
        # vehicles
        v_ct, v_moff, _ = read_reflexive(map_data, moff + 32 - magic, 20, 16, magic)
        repair_dependency_array(*(args + (b'ejbo', v_moff, v_ct)))
        # shaders
        repair_dependency_array(*(args + (b'rdhs', moff + 44, 2)))
        repair_dependency(*(args + (b'meti', moff + 76)))
        # sounds
        s_ct, s_moff, _ = read_reflexive(map_data, moff + 92 - magic, 60, 16, magic)
        repair_dependency_array(*(args + (b'!dns', s_moff, s_ct)))

    # player info
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x170 - magic, 244, 1, magic):
        repair_dependency(*(args + (b'tinu', moff)))
        repair_dependency(*(args + (b'effe', moff + 184)))

    # fp interface
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x17C - magic, 192, 1, magic):
        repair_dependency(*(args + (b'2dom', moff)))
        repair_dependency(*(args + (b'mtib', moff + 16)))
        repair_dependency(*(args + (b'rtem', moff + 32)))
        repair_dependency(*(args + (b'rtem', moff + 52)))
        repair_dependency_array(*(args + (b'effe', moff + 72, 2)))

    # falling damage
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x188 - magic, 152, 1, magic):
        repair_dependency(*(args + (b'!tpj', moff + 16)))
        repair_dependency_array(*(args + (b'!tpj', moff + 44, 5)))

    # materials
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x194 - magic, 884, MAX_MATERIAL_COUNT, magic):
        repair_dependency(*(args + (b'effe', moff + 740)))
        repair_dependency(*(args + (b'!dns', moff + 756)))
        repair_dependency(*(args + (b'!dns', moff + 868)))
        # particle effects
        p_ct, p_moff, _ = read_reflexive(map_data, moff + 796 - magic, 8, 128, magic)
        repair_dependency_array(*(args + (b'trap', p_moff, p_ct, 128)))


def repair_mgs2(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'mtib', index_array[tag_id].meta_offset + 92)


def repair_mode(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    # shaders
    shader_ct, shader_moff, _ = read_reflexive(
        map_data, tag_offset + 220 - magic, 256, 32, magic)

    if safe_mode:
        used_shader_indices = set()
        # loop over each geometries parts and determine which
        # shader indices are actually used across all of them
        for moff in iter_reflexive_offs(map_data, tag_offset + 208 - magic, 48):
            for moff2 in iter_reflexive_offs(map_data, moff + 36 - magic, 132):
                map_data.seek(moff2 + 4 - magic)
                shader_type = int.from_bytes(
                    map_data.read(2), 'little', signed=True)
                if shader_type in range(shader_ct):
                    used_shader_indices.add(shader_type)
    else:
        used_shader_indices = list(range(shader_ct))

    for i in sorted(used_shader_indices):
        repair_dependency(index_array, map_data, magic, repair, engine,
                          b'rdhs', shader_moff + 32 * i)


def repair_mply(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # scenario descriptions
    for moff in iter_reflexive_offs(map_data, tag_offset - magic, 68, 32, magic):
        repair_dependency(*(args + (b'mtib', moff)))
        repair_dependency(*(args + (b'rtsu', moff + 16)))


def repair_ngpr(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'mtib', tag_offset + 56)))
    repair_dependency(*(args + (b'mtib', tag_offset + 76)))


def repair_predicted_resources(map_data, offset, magic, repair, max_count=0xFFffFFff):
    for moff in iter_reflexive_offs(map_data, offset - magic, 8, max_count, magic):
        map_data.seek(moff - magic)
        rsrc_type = map_data.read(4)[:2]
        tag_id = int.from_bytes(map_data.read(4), "little") & 0xFFFF

        if tag_id == 0xFFFF:
            continue
        elif rsrc_type == b'\x00\x00':
            # bitmap resource type
            repair[tag_id] = 'bitm'
        elif rsrc_type == b'\x01\x00':
            # sound resource type
            repair[tag_id] = 'snd!'


def repair_object(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    map_data.seek(tag_offset - magic)
    object_type = int.from_bytes(map_data.read(2), 'little')
    if object_type != -1 and object_type not in range(len(object_class_bytes) - 1):
        # not an object
        return

    # obje_attrs struct size is 380
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'2dom', tag_offset + 40)))
    repair_dependency(*(args + (b'rtna', tag_offset + 56)))

    repair_dependency(*(args + (b'lloc', tag_offset + 112)))
    repair_dependency(*(args + (b'syhp', tag_offset + 128)))
    repair_dependency(*(args + (b'rdhs', tag_offset + 144)))
    repair_dependency(*(args + (b'effe', tag_offset + 160)))

    # attachments
    ct, moff, _ = read_reflexive(map_data, tag_offset + 320 - magic, 8, 72, magic)
    repair_dependency_array(*(args + (None, moff, ct, 72)))

    # widgets
    ct, moff, _ = read_reflexive(map_data, tag_offset + 332 - magic, 4, 32, magic)
    repair_dependency_array(*(args + (None, moff, ct, 32)))

    # add the predicted resources to the tags to repair
    repair_predicted_resources(map_data, tag_offset + 368, magic, repair, 1024)

    tag_offset += 380
    args2 = (index_array, map_data, magic, repair, engine)
    if object_type <= 1:
        # bipd or vehi
        repair_unit_attrs(tag_offset, *args2)
        tag_offset += 372
        if object_type == 0:
            # bipd
            repair_dependency(*(args + (b'!tpj', tag_offset + 36)))
            repair_dependency(*(args + (b'toof', tag_offset + 156)))
        else:
            # vehi
            repair_dependency(*(args + (b'!dns', tag_offset + 192)))
            repair_dependency(*(args + (b'!dns', tag_offset + 208)))
            repair_dependency(*(args + (b'toof', tag_offset + 224)))
            repair_dependency(*(args + (b'effe', tag_offset + 240)))

    elif object_type <= 4:
        # weap, eqip, or garb
        repair_item_attrs(tag_offset, *args2)
        tag_offset += 396
        if object_type == 2:
            # weap
            repair_dependency(*(args + (None, tag_offset + 52)))
            repair_dependency(*(args + (None, tag_offset + 108)))
            repair_dependency(*(args + (None, tag_offset + 124)))
            repair_dependency(*(args + (b'!tpj', tag_offset + 140)))
            repair_dependency(*(args + (b'!tpj', tag_offset + 156)))
            repair_dependency(*(args + (b'vtca', tag_offset + 180)))
            repair_dependency(*(args + (None, tag_offset + 280)))
            repair_dependency(*(args + (None, tag_offset + 296)))
            repair_dependency(*(args + (b'2dom', tag_offset + 340)))
            repair_dependency(*(args + (b'rtna', tag_offset + 356)))
            repair_dependency(*(args + (b'ihpw', tag_offset + 376)))
            repair_dependency(*(args + (b'!dns', tag_offset + 392)))
            repair_dependency(*(args + (b'!dns', tag_offset + 408)))
            repair_dependency(*(args + (b'!dns', tag_offset + 424)))

            repair_predicted_resources(
                map_data, tag_offset + 476, magic, repair, 1024)

            # magazines
            for moff in iter_reflexive_offs(
                    map_data, tag_offset + 488 - magic, 112, 2, magic):
                repair_dependency(*(args + (None, moff + 56)))
                repair_dependency(*(args + (None, moff + 72)))

                # magazine items
                ct, moff2, _ = read_reflexive(map_data, moff + 100 - magic, 2, 28, magic)
                repair_dependency_array(*(args + (b'piqe', moff2 + 12, ct, 28)))

            # triggers
            for moff in iter_reflexive_offs(
                    map_data, tag_offset + 500 - magic, 276, 2, magic):
                repair_dependency(*(args + (None, moff + 92)))
                repair_dependency(*(args + (b'ejbo', moff + 148)))

                # firing effects
                for moff2 in iter_reflexive_offs(
                        map_data, moff + 264 - magic, 132, 8, magic):
                    repair_dependency(*(args + (None, moff2 + 36)))
                    repair_dependency(*(args + (None, moff2 + 52)))
                    repair_dependency(*(args + (None, moff2 + 68)))
                    repair_dependency(*(args + (b'!tpj', moff2 + 84)))
                    repair_dependency(*(args + (b'!tpj', moff2 + 100)))
                    repair_dependency(*(args + (b'!tpj', moff2 + 116)))

        elif object_type == 3:
            # eqip
            repair_dependency(*(args + (b'!dns', tag_offset + 8)))
        else:
            # garb
            pass  # nothing else to do for this

    elif object_type == 5:
        # proj
        repair_dependency(*(args + (b'effe', tag_offset + 16)))
        repair_dependency(*(args + (b'effe', tag_offset + 48)))
        repair_dependency(*(args + (b'effe', tag_offset + 120)))
        repair_dependency(*(args + (b'!dns', tag_offset + 136)))
        repair_dependency(*(args + (b'!tpj', tag_offset + 152)))
        repair_dependency(*(args + (b'!tpj', tag_offset + 168)))

        # material responses
        for moff in iter_reflexive_offs(
                map_data, tag_offset + 196 - magic, 160, MAX_MATERIAL_COUNT, magic):
            repair_dependency(*(args + (b'effe', moff + 4)))
            repair_dependency(*(args + (b'effe', moff + 60)))
            repair_dependency(*(args + (b'effe', moff + 104)))
    elif object_type == 6:
        # scen
        pass  # nothing else to do for this

    elif object_type <= 9:
        # mach, ctrl, or lifi
        repair_devi_attrs(tag_offset, *args2)
        tag_offset += 276
        if object_type == 8:
            # ctrl
            repair_dependency(*(args + (None, tag_offset + 88)))
            repair_dependency(*(args + (None, tag_offset + 104)))
            repair_dependency(*(args + (None, tag_offset + 120)))
        else:
            # mach, lifi
            pass  # nothing else to do for these

    else:
        # plac, or ssce
        pass  # nothing else to do for these


def repair_part(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'mtib', tag_offset + 0x4)))
    repair_dependency(*(args + (b'yhpp', tag_offset + 0x14)))
    repair_dependency(*(args + (b'toof', tag_offset + 0x24)))
    repair_dependency(*(args + (None, tag_offset + 0x48)))
    repair_dependency(*(args + (None, tag_offset + 0x58)))
    repair_dependency(*(args + (b'mtib', tag_offset + 0xFC)))


def repair_pctl(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'yhpp', tag_offset + 56)))

    # particle types
    for moff in iter_reflexive_offs(map_data, tag_offset + 92 - magic, 128, 4, magic):
        # particle states
        for moff2 in iter_reflexive_offs(map_data, moff + 116 - magic, 376, 8, magic):
            repair_dependency(*(args + (b'mtib', moff2 + 48)))
            repair_dependency(*(args + (b'yhpp', moff2 + 132)))
            repair_dependency(*(args + (b'mtib', moff2 + 260)))


def repair_rain(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # particle types
    for moff in iter_reflexive_offs(map_data, tag_offset + 0x24 - magic, 0x25C, 8, magic):
        repair_dependency(*(args + (b'yhpp', moff + 0xAC)))
        repair_dependency(*(args + (b'mtib', moff + 0x194)))
        repair_dependency(*(args + (b'mtib', moff + 0x1F4)))


def repair_sbsp(tag_offset, index_array, map_data, magic, repair, engine,
                map_magic, safe_mode=True):
    # This function requires the first argument is the tag's magic offset
    # relative to the bsp magic, rather than the tag's index id
    args = (index_array, map_data, magic, repair, engine)
    kwargs = dict(map_magic=map_magic)
    repair_dependency(*(args + (b'mtib', tag_offset)), **kwargs)

    # collision materials
    ct, moff, _ = read_reflexive(map_data, tag_offset + 164 - magic, 512, 20, magic)
    repair_dependency_array(*(args + (b'rdhs', moff, ct, 20)), **kwargs)

    # lightmaps
    for moff in iter_reflexive_offs(map_data, tag_offset + 260 - magic, 32, 128, magic):
        ct, moff2, _ = read_reflexive(map_data, moff + 20 - magic, 2048, 256, magic)
        # materials
        repair_dependency_array(*(args + (b'rdhs', moff2, ct, 256)), **kwargs)

    # lens flares
    ct, moff, _ = read_reflexive(map_data, tag_offset + 284 - magic, 256, 16, magic)
    repair_dependency_array(*(args + (b'snel', moff, ct)), **kwargs)

    # fog palettes
    ct, moff, _ = read_reflexive(map_data, tag_offset + 400 - magic, 32, 136, magic)
    repair_dependency_array(*(args + (b' gof', moff + 32, ct, 136)), **kwargs)

    # weather palettes
    for moff in iter_reflexive_offs(map_data, tag_offset + 436 - magic, 240, 32, magic):
        repair_dependency(*(args + (b'niar', moff + 32)), **kwargs)
        repair_dependency(*(args + (b'dniw', moff + 128)), **kwargs)

    # background sounds palette
    ct, moff, _ = read_reflexive(map_data, tag_offset + 508 - magic, 64, 116, magic)
    repair_dependency_array(*(args + (b'dnsl', moff + 32, ct, 116)), **kwargs)

    # sound environments palette
    ct, moff, _ = read_reflexive(map_data, tag_offset + 520 - magic, 64, 80, magic)
    repair_dependency_array(*(args + (b'edns', moff + 32, ct, 80)), **kwargs)


def repair_scnr(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    ### Need to finish this up. not all the limits specified here
    # should be as low as they are because open sauce is a thing
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    if "yelo" in engine:
        repair_dependency(*(args + (b'oley', tag_offset)))

        # bsp modifiers
        for moff in iter_reflexive_offs(
                map_data, tag_offset + 1288 - magic, 64, 32, magic):
            # lightmap sets
            for moff2 in iter_reflexive_offs(map_data, moff + 4 - magic,
                                             124, 64, magic):
                repair_dependency_array(*(args + (b'mtib', moff2 + 36, 4)))

            # sky set
            for moff2 in iter_reflexive_offs(map_data, moff + 16 - magic, 44, 64, magic):
                # skies
                for moff3 in iter_reflexive_offs(map_data, moff2 + 32 - magic, 20, 8, magic):
                    repair_dependency(*(args + (b' yks', moff3 + 4)))

    repair_dependency(*(args + (b'rtsu', tag_offset + 1396)))
    repair_dependency(*(args + (b'rtsu', tag_offset + 1412)))
    repair_dependency(*(args + (b' tmh', tag_offset + 1428)))

    # skies
    ct, moff, _ = read_reflexive(map_data, tag_offset + 48 - magic, 8, 16, magic)
    repair_dependency_array(*(args + (b' yks', moff, ct)))

    # player starting profiles
    for moff in iter_reflexive_offs(map_data, tag_offset + 840 - magic, 104, 256, magic):
        repair_dependency(*(args + (b'paew', moff + 40)))
        repair_dependency(*(args + (b'paew', moff + 60)))

    # netgame flags
    ct, moff, _ = read_reflexive(map_data, tag_offset + 888 - magic, 200, 148, magic)
    repair_dependency_array(*(args + (b'cmti', moff + 20, ct, 148)))

    # netgame equipments
    ct, moff, _ = read_reflexive(map_data, tag_offset + 900 - magic, 200, 144, magic)
    repair_dependency_array(*(args + (b'cmti', moff + 80, ct, 144)))

    # starting equipment
    for moff in iter_reflexive_offs(map_data, tag_offset + 912 - magic, 204, 200, magic):
        repair_dependency(*(args + (b'cmti', moff + 60)))
        repair_dependency(*(args + (b'cmti', moff + 76)))
        repair_dependency(*(args + (b'cmti', moff + 92)))
        repair_dependency(*(args + (b'cmti', moff + 108)))
        repair_dependency(*(args + (b'cmti', moff + 124)))
        repair_dependency(*(args + (b'cmti', moff + 140)))

    # ai animation reference
    ct, moff, _ = read_reflexive(map_data, tag_offset + 1092 - magic, 128, 60, magic)
    repair_dependency_array(*(args + (b'rtna', moff + 32, ct, 60)))

    # ai conversations
    for moff in iter_reflexive_offs(map_data, tag_offset + 1128 - magic, 116, 128, magic):
        # lines
        for moff2 in iter_reflexive_offs(map_data, moff + 92 - magic, 124, 32, magic):
            repair_dependency(*(args + (b'!dns', moff2 + 28)))
            repair_dependency(*(args + (b'!dns', moff2 + 44)))
            repair_dependency(*(args + (b'!dns', moff2 + 60)))
            repair_dependency(*(args + (b'!dns', moff2 + 76)))
            repair_dependency(*(args + (b'!dns', moff2 + 92)))
            repair_dependency(*(args + (b'!dns', moff2 + 108)))

    # tag references
    ct, moff, _ = read_reflexive(map_data, tag_offset + 1204 - magic, 256, 40, magic)
    repair_dependency_array(*(args + (None, moff + 24, ct, 40)))

    # structure bsps
    ct, moff, _ = read_reflexive(map_data, tag_offset + 1444 - magic, 32, 32, magic)
    repair_dependency_array(*(args + (b'psbs', moff + 16, ct, 32)))

    # palettes
    # NOTE: Can't trust that these palettes are valid.
    # Need to check what the highest one used by all instances
    for off, inst_size in (
            (540, 72), (564, 120), (588, 120), # scen  bipd  vehi
            (612, 40), (636, 92), (672, 64),   # eqip  weap  mach
            (696, 64), (720, 88), (744, 40)):  # ctrl  lifi  ssce
        pal_ct, pal_moff, _ = read_reflexive(map_data, tag_offset + off - magic)

        if safe_mode:
            used_pal_indices = set()
            # loop over each object instance and determine which
            # palette indices are actually used across all of them
            for moff in iter_reflexive_offs(
                    map_data, tag_offset + off - 12 - magic, inst_size, tag_magic=magic):
                map_data.seek(moff - magic)
                inst_type = int.from_bytes(map_data.read(2), 'little', signed=True)
                if inst_type in range(pal_ct):
                    used_pal_indices.add(inst_type)
        else:
            used_pal_indices = list(range(pal_ct))

        for i in sorted(used_pal_indices):
            repair_dependency(*(args + (b'ejbo', pal_moff + 48 * i)))

    # script syntax data references
    size, _, __, moff, ___ = read_rawdata_ref(map_data, tag_offset + 1140 - magic, magic)
    map_data.seek(moff - magic)
    script_syntax_data_nodes = get_hsc_data_block(map_data.read(size)).nodes
    for node in script_syntax_data_nodes:
        tag_cls = {
            24: 'snd!', 25: 'effe', 26: 'jpt!', 27: 'lsnd',
            28: 'antr', 29: 'actv', 30: 'jpt!', 31: 'obje'
            }.get(node.type)
        if tag_cls is None or (node.flags & HSC_IS_SCRIPT_OR_GLOBAL):
            continue

        sub_tag_id = node.data & 0xFFff
        if sub_tag_id in repair or sub_tag_id not in range(len(index_array)):
            continue

        if tag_cls == "obje":
            try:
                map_data.seek(index_array[sub_tag_id].meta_offset - magic)
                object_type = int.from_bytes(map_data.read(2), 'little')
                if object_type not in range(-1, len(object_class_bytes) - 1):
                    continue
            except Exception:
                continue

            tag_cls = object_class_bytes[object_type]

        repair[sub_tag_id] = tag_cls


    # decals
    ct, moff, _ = read_reflexive(map_data, tag_offset + 948 - magic, 128, 16, magic)
    repair_dependency_array(*(args + (b'aced', moff, ct)))

    # detail objects
    ct, moff, _ = read_reflexive(map_data, tag_offset + 960 - magic, 32, 48, magic)
    repair_dependency_array(*(args + (b'cbod', moff, ct, 48)))

    # actors palette
    ct, moff, _ = read_reflexive(map_data, tag_offset + 1056 - magic, 64, 16, magic)
    repair_dependency_array(*(args + (b'vtca', moff, ct)))


def repair_shader(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    map_data.seek(tag_offset + 36 - magic)
    shader_type = int.from_bytes(map_data.read(2), 'little')

    if shader_type != -1 and shader_type not in range(len(shader_class_bytes) - 1):
        # not a shader
        return

    typ = shader_class_bytes[shader_type]

    tag_offset += 40
    args = (index_array, map_data, magic, repair, engine)

    if typ == b'vnes':
        repair_dependency(*(args + (b'snel', tag_offset + 0x8)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x60)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x90)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0xA4)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0xD4)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x100)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x22C)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x2FC)))
        # shader environment os extension
        if "yelo" in engine:
            ct, moff, _ = read_reflexive(map_data, tag_offset + 0xC8 - magic, 1, 100, magic)
            repair_dependency_array(*(args + (b'mtib', moff + 8, ct, 100)))

    elif typ == b'osos':
        repair_dependency(*(args + (b'mtib', tag_offset + 0x7C)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x94)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0xB4)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x13C)))
        # shader model os extension
        if "yelo" in engine:
            for moff in iter_reflexive_offs(
                    map_data, tag_offset + 0xC8 - magic, 192, 1, magic):
                repair_dependency(*(args + (b'mtib', moff)))
                repair_dependency(*(args + (b'mtib', moff + 0x20)))
                repair_dependency(*(args + (b'mtib', moff + 0x40)))
                repair_dependency(*(args + (b'mtib', moff + 0x60)))

    elif typ == b'tems':
        repair_dependency(*(args + (b'mtib', tag_offset + 0x24)))

    elif typ == b'algs':
        repair_dependency(*(args + (b'mtib', tag_offset + 0x3C)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x84)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x98)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x130)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x144)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x178)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x18C)))

    elif typ == b'alps':
        repair_dependency(*(args + (b'mtib', tag_offset + 0xAC)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0xF4)))

    elif typ == b'taws':
        repair_dependency(*(args + (b'mtib', tag_offset + 0x24)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0x74)))
        repair_dependency(*(args + (b'mtib', tag_offset + 0xA0)))

    elif typ in b'rtos_ihcs_xecs':
        repair_dependency(*(args + (b'snel', tag_offset + 0x10)))
        # layers
        ct, moff, _ = read_reflexive(map_data, tag_offset + 0x20 - magic, 4, 16, magic)
        repair_dependency_array(*(args + (b'rdhs', moff, ct)))
        # maps
        maps_size = 0x64 if typ == b'rtos' else 0xDC

        ct, moff, _ = read_reflexive(map_data, tag_offset + 0x2C - magic, 4, maps_size, magic)

        moff += 0x1C if typ == b'rtos' else 0x6C

        repair_dependency_array(*(args + (b'mtib', moff, ct, maps_size)))

        if typ == b'xecs':
            # 2 stage maps
            ct, moff, _ = read_reflexive(map_data, tag_offset + 0x38 - magic, 2, 0xDC, magic)
            repair_dependency_array(*(args + (b'mtib', moff + 0x6C, ct, 0xDC)))


def repair_sky(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'2dom', tag_offset)))
    repair_dependency(*(args + (b'rtna', tag_offset + 0x10)))
    repair_dependency(*(args + (b' gof', tag_offset + 0x98)))
    # lights
    ct, moff, _ = read_reflexive(map_data, tag_offset + 0xC4 - magic, 8, 116, magic)
    repair_dependency_array(*(args + (b'snel', moff, ct, 116)))


def repair_snd_(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'!dns', index_array[tag_id].meta_offset + 0x70)


def repair_Soul(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_tagc(tag_id, index_array, map_data, magic, repair, engine, safe_mode, b'aLeD', 32)


def repair_tagc(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True,
                tag_cls=None, max_count=200):
    ct, moff, _ = read_reflexive(
        map_data, index_array[tag_id].meta_offset - magic, max_count, 16, magic)

    for moff2 in range(moff, moff + (ct * 16), 16):
        map_data.seek(moff2 + 12 - magic)
        tag_id = int.from_bytes(map_data.read(4), "little")
        if tag_id == 0xFFffFFff:
            continue
        elif ((tag_id & 0xFFff) not in range(len(index_array)) or
              index_array[tag_id & 0xFFff].id != tag_id):
            # break on the first invalid tag id.
            # the reflexive size might be corrupt
            break

        repair_dependency(index_array, map_data, magic, repair, engine,
                          tag_cls, moff2)


def repair_udlg(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine, b'!dns')
    repair_dependency_array(*(args + (tag_offset + 0x10, 3)))
    repair_dependency_array(*(args + (tag_offset + 0x70, 14)))
    repair_dependency_array(*(args + (tag_offset + 0x160, 4)))
    repair_dependency_array(*(args + (tag_offset + 0x1E0, 17)))
    repair_dependency_array(*(args + (tag_offset + 0x320, 28)))
    repair_dependency_array(*(args + (tag_offset + 0x510, 13)))
    repair_dependency_array(*(args + (tag_offset + 0x610, 10)))
    repair_dependency_array(*(args + (tag_offset + 0x6D0, 13)))
    repair_dependency_array(*(args + (tag_offset + 0x7C0, 21)))
    repair_dependency_array(*(args + (tag_offset + 0x950, 23)))
    repair_dependency_array(*(args + (tag_offset + 0xB20, 7)))
    repair_dependency_array(*(args + (tag_offset + 0xBD0, 5)))
    repair_dependency_array(*(args + (tag_offset + 0xC60, 8)))


def repair_unhi(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # unit hud background
    repair_hud_background(*(args + (tag_offset + 36, )))
    # shield panel background
    repair_hud_background(*(args + (tag_offset + 140, )))
    # shield panel meter
    repair_dependency(*(args + (b'mtib', tag_offset + 280)))
    # health panel background
    repair_hud_background(*(args + (tag_offset + 380, )))
    # health panel meter
    repair_dependency(*(args + (b'mtib', tag_offset + 520)))
    # motion sensor panel background
    repair_hud_background(*(args + (tag_offset + 620, )))
    # motion sensor panel foreground
    repair_hud_background(*(args + (tag_offset + 724, )))

    # auxilary overlay
    for moff in iter_reflexive_offs(map_data, tag_offset + 932 - magic, 132, 16, magic):
        repair_hud_background(*(args + (moff, )))

    # warning sounds
    for moff in iter_reflexive_offs(map_data, tag_offset + 960 - magic, 56, 12, magic):
        map_data.seek(moff - magic)
        if map_data.read(4) == b'dnsl':
            repair_dependency(*(args + (b'dnsl', moff)))
        else:
            repair_dependency(*(args + (b'!dns', moff)))

    # auxilary meter
    for moff in iter_reflexive_offs(map_data, tag_offset + 972 - magic, 324, 16, magic):
        repair_hud_background(*(args + (moff + 20, )))
        repair_dependency(*(args + (b'mtib', moff + 160)))


def repair_vcky(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'tnof', tag_offset)))
    repair_dependency(*(args + (b'mtib', tag_offset + 16)))
    repair_dependency(*(args + (b'rtsu', tag_offset + 32)))
    # virtual keys
    for moff in iter_reflexive_offs(map_data, tag_offset + 48 - magic, 80, 44, magic):
        repair_dependency_array(*(args + (b'mtib', moff + 16, 4)))


def repair_wphi(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'ihpw', tag_offset)))

    # static elements
    for moff in iter_reflexive_offs(map_data, tag_offset + 96 - magic, 180, 16, magic):
        repair_dependency(*(args + (b'mtib', moff + 72)))
        # multitex overlays
        for moff2 in iter_reflexive_offs(map_data, moff + 124 - magic, 480, 30, magic):
            repair_dependency(*(args + (b'mtib', moff2 + 100)))
            repair_dependency(*(args + (b'mtib', moff2 + 116)))
            repair_dependency(*(args + (b'mtib', moff2 + 132)))

    # meter elements
    for moff in iter_reflexive_offs(map_data, tag_offset + 108 - magic, 180, 16, magic):
        repair_dependency(*(args + (b'mtib', moff + 72)))

    # crosshairs
    for moff in iter_reflexive_offs(map_data, tag_offset + 132 - magic, 104, 19, magic):
        repair_dependency(*(args + (b'mtib', moff + 36)))

    # overlay elements
    for moff in iter_reflexive_offs(map_data, tag_offset + 144 - magic, 104, 16, magic):
        repair_dependency(*(args + (b'mtib', moff + 36)))

    # screen effects
    for moff in iter_reflexive_offs(map_data, tag_offset + 172 - magic, 184, 1, magic):
        repair_dependency(*(args + (b'mtib', moff + 24)))
        repair_dependency(*(args + (b'mtib', moff + 40)))


# open-sauce repair functions
def repair_avtc(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # actor variant transforms
    for moff in iter_reflexive_offs(map_data, tag_offset - magic, 52, 32, magic):
        repair_dependency(*(args + (b'vtca', moff)))

        # transforms
        for moff2 in iter_reflexive_offs(map_data, moff + 16 - magic, 116, 32, magic):
            repair_dependency(*(args + (b'otva', moff2 + 52)))
            repair_dependency(*(args + (b'itva', moff2 + 72)))


def repair_avti(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    # targets
    for moff in iter_reflexive_offs(map_data, tag_offset - magic, 172, 16, magic):
        repair_dependency(*(args + (b'vtca', moff + 52)))

        # keyframe actions
        for moff2 in iter_reflexive_offs(map_data, moff + 120 - magic, 72, 9, magic):
            repair_dependency(*(args + (b'!tpj', moff2 + 8)))
            repair_dependency(*(args + (b'effe', moff2 + 24)))


def repair_avto(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset - magic
    args = (index_array, map_data, magic, repair, engine)

    # instigators
    for moff in iter_reflexive_offs(map_data, tag_offset + 44, 32, 16, magic):
        repair_dependency(*(args + (b'tinu', moff)))

    # keyframe actions
    for moff in iter_reflexive_offs(map_data, tag_offset + 88, 72, 9, magic):
        repair_dependency(*(args + (b'!tpj', moff + 8)))
        repair_dependency(*(args + (b'effe', moff + 24)))

    # attachments
    for moff in iter_reflexive_offs(map_data, tag_offset + 104, 120, 16, magic):
        repair_dependency(*(args + (b'ejbo', moff)))


def repair_efpc(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    # effects
    ct, moff, _ = read_reflexive(
        map_data, index_array[tag_id].meta_offset + 24 - magic, 32, 72, magic)
    repair_dependency_array(
        index_array, map_data, magic, repair, engine, b'gpfe', moff, ct, 72)


def repair_efpg(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    # shaders
    ct, moff, _ = read_reflexive(
        map_data, index_array[tag_id].meta_offset + 60 - magic, 12, 16, magic)
    repair_dependency_array(
        index_array, map_data, magic, repair, engine, b'gphs', moff, ct)


def repair_gelc(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    # unit infections
    for moff in iter_reflexive_offs(map_data, tag_offset + 4 - magic, 52, 1, magic):
        ct, moff2, _ = read_reflexive(map_data, moff + 4 - magic, 16, magic)
        repair_dependency_array(*(args + (b'tinu', moff2, ct)))

        # boarding seats
        for moff2 in iter_reflexive_offs(map_data, moff + 16 - magic, 144, 32, magic):
            repair_dependency(*(args + (b'tinu', moff2 + 4)))
            repair_dependency(*(args + (b'tinu', moff2 + 24)))
            repair_dependency(*(args + (b'vtca', moff2 + 40)))
            repair_dependency(*(args + (b'effe', moff2 + 56)))
            repair_dependency(*(args + (b'ejbo', moff2 + 72)))

    # unit external upgrades
    for moff in iter_reflexive_offs(map_data, tag_offset + 16 - magic, 68, 64, magic):
        repair_dependency(*(args + (b'tinu', moff + 4)))

        ct, moff2, _ = read_reflexive(map_data, moff + 20 - magic, 120, magic)
        repair_dependency_array(*(args + (b'!tpj', moff2 + 72, ct, 120)))


def repair_gelo(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'cgat', tag_offset + 40)))
    repair_dependency(*(args + (b'cleg', tag_offset + 56)))

    # scripted ui widgets
    ct, moff, _ = read_reflexive(map_data, tag_offset + 152 - magic, 128, 76, magic)
    repair_dependency_array(*(args + (b'aLeD', moff + 32, ct, 76)))


def repair_magy(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_antr(tag_id, index_array, map_data, magic, repair, engine)
    repair_dependency(index_array, map_data, magic, repair, engine,
                      b'rtna', index_array[tag_id].meta_offset + 0x80)


def repair_shpg(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_shpp(tag_id, index_array, map_data, magic, repair, engine)

    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)
    repair_dependency(*(args + (b'gphs', tag_offset + 168)))

    # merged values
    for moff in iter_reflexive_offs(map_data, tag_offset + 184 - magic, 116, 16*8, magic):
        repair_dependency(*(args + (b'mtib', moff + 88)))


def repair_shpp(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    repair_predicted_resources(
        map_data, index_array[tag_id].meta_offset + 120, magic, repair, 1024)


def repair_unic(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    ct, moff, _ = read_reflexive(
        map_data, index_array[tag_id].meta_offset - magic, 9216, 56, magic)
    repair_dependency_array(index_array, map_data, magic, repair, engine,
                            b'ydis', moff, ct, 56)


def repair_yelo(tag_id, index_array, map_data, magic, repair, engine, safe_mode=True):
    tag_offset = index_array[tag_id].meta_offset
    args = (index_array, map_data, magic, repair, engine)

    repair_dependency(*(args + (b'oleg', tag_offset + 4)))
    repair_dependency(*(args + (b'gtam', tag_offset + 20)))
    repair_dependency(*(args + (b'cgat', tag_offset + 36)))

    # scripted ui widgets
    ct, moff, _ = read_reflexive(map_data, tag_offset + 104 - magic, 128, 76, magic)
    repair_dependency_array(*(args + (b'aLeD', moff + 32, ct, 76)))


class_repair_functions = {
    "actv": repair_actv, "ant!": repair_ant_, "antr": repair_antr,
    "coll": repair_coll, "cont": repair_cont, "deca": repair_deca,
    "DeLa": repair_DeLa, "dobc": repair_dobc, "effe": repair_effe,
    "elec": repair_elec, "flag": repair_flag, "fog ": repair_fog,
    "font": repair_font, "foot": repair_foot, "glw!": repair_glw_,
    "grhi": repair_grhi, "hud#": repair_hud_, "hudg": repair_hudg,
    "itmc": repair_itmc, "jpt!": repair_jpt_, "lens": repair_lens,
    "ligh": repair_ligh, "lsnd": repair_lsnd, "matg": repair_matg,
    "mgs2": repair_mgs2, "mode": repair_mode, "mod2": repair_mode,
    "mply": repair_mply, "ngpr": repair_ngpr, "part": repair_part,
    "pctl": repair_pctl, "rain": repair_rain, "sbsp": repair_sbsp,
    "scnr": repair_scnr, "sky ": repair_sky,  "snd!": repair_snd_,
    "Soul": repair_Soul, "tagc": repair_tagc, "udlg": repair_udlg,
    "unhi": repair_unhi, "vcky": repair_vcky, "wphi": repair_wphi,

    # open sauce
    "avtc": repair_avtc, "avti": repair_avti, "avto": repair_avto,
    "efpc": repair_efpc, "efpg": repair_efpg, "gelc": repair_gelc,
    "gelo": repair_gelo, "magy": repair_magy, "shpg": repair_shpg,
    "shpp": repair_shpp, "unic": repair_unic, "yelo": repair_yelo
    }

# object subclasses
for name in ("bipd", "vehi", "weap", "eqip", "garb", "proj", "scen",
             "mach", "ctrl", "lifi", "plac", "ssce", "obje"):
    class_repair_functions[name] = repair_object

# shader subclasses
for name in ("senv", "soso", "sotr", "schi", "scex", "swat", "sgla",
             "smet", "spla", "shdr"):
    class_repair_functions[name] = repair_shader

# make a copy of the class_repair_functions, but have the
# functions indexed by the reversed fcc string as bytes
_class_repair_functions_by_bytes = {
    bytes(k[slice(None, None, -1)], "latin1"): class_repair_functions[k]
    for k in class_repair_functions}



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

soso_comment = """MODEL SHADER
Setting <true atmospheric fog> enables per-pixel atmospheric fog but disables point/spot
lights, planar fog, and the ability to control the atmospheric fog density for this shader."""

cc_comment = """CHANGE COLOR
Change color is used to recolor the diffuse map, it affects pixels
based on the ALPHA channel (BLUE on XBOX) of the multipurpose map."""

self_illum_comment = """SELF-ILLUMINATION
Self-illumination adds diffuse light to pixels based on the GREEN channel
of the multipurpose map. The external self-illumination color referenced by
<color source> is modulated by the self-illumination animation."""

maps_comment = """MAPS
On PC, the multipurpose map channels are used for:
* RED: auxiliary mask (usually used for detail)
* GREEN: self-illumination mask (adds to diffuse light)
* BLUE: specular reflection mask (modulates reflections)
* ALPHA: color change mask (recolors diffuse map)

On XBOX, the channels are used for:
* RED: specular reflection
* GREEN: self-illumination
* BLUE: color change
* ALPHA: auxiliary

Note: When DXT1 compressed color-key textures are used for the
multipurpose map (as they should be normally), the alpha channel is 1-bit
and any non-zero alpha pixels must have zero-color, therefore on PC if we
need colorchange we use DXT3 (explicit alpha) or DXT5 (interpolated alpha).

Detail map affects diffuse map, and optionally affects reflection
if <detail after reflection> flag is set."""

tex_scroll_comment = """TEXTURE SCROLLING ANIMATIONS
Scrolls all 2D maps simultaneously."""

reflection_prop_comment = """REFLECTION PROPERTIES"""

model_shader = Struct("model_shader",
    Bool16("flags",
        "detail_after_reflection",
        "two_sided",
        "not_alpha_tested",
        "alpha_blended_decal",
        "true_atmospheric_fog",
        "disable_two_sided_culling",
        ),
    Pad(14),
    Float("translucency"),
    COMMENT=soso_comment
    )

self_illumination = Struct("self_illumination",
    Bool16("flags",
        "no_random_phase"
        ),
    Pad(2),
    SEnum16("color_source", *function_names),
    SEnum16("animation_function", *animation_functions),
    float_sec("animation_period"),  # seconds
    QStruct("color_lower_bound", INCLUDE=rgb_float),
    QStruct("color_upper_bound", INCLUDE=rgb_float),
    COMMENT=self_illum_comment
    )

maps = Struct("maps",
    Float("map_u_scale"),
    Float("map_v_scale"),
    dependency("diffuse_map", "bitm"),

    Pad(8),
    dependency("multipurpose_map", "bitm"),

    Pad(8),
    SEnum16("detail_function", *detail_map_functions),
    SEnum16("detail_mask", *detail_mask),

    Float("detail_map_scale"),
    dependency("detail_map", "bitm"),
    Float("detail_map_v_scale"),
    COMMENT=maps_comment
    )

texture_scrolling = Struct("texture_scrolling",
    Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),
    QStruct("rot_animation_center", INCLUDE=xy_float),
    COMMENT=tex_scroll_comment
    )

reflection = Struct("reflection",
    float_wu("falloff_distance"),  # world units
    float_wu("cutoff_distance"),  # world units

    float_zero_to_one("perpendicular_brightness"),
    QStruct("perpendicular_tint_color", INCLUDE=rgb_float),
    float_zero_to_one("parallel_brightness"),
    QStruct("parallel_tint_color", INCLUDE=rgb_float),

    dependency("cube_map", "bitm"),
    #COMMENT=reflection_prop_comment
    )

soso_attrs = Struct("soso_attrs",
    #Model Shader Properties
    model_shader,

    Pad(16),
    #Color-Change
    SEnum16("color_change_source", *function_names, COMMENT=cc_comment),


    Pad(30),
    #Self-Illumination
    self_illumination,

    Pad(12),
    #Diffuse, Multipurpose, and Detail Maps
    maps,

    # this padding is the reflexive for the OS shader model extension
    Pad(12),

    #Texture Scrolling Animation
    texture_scrolling,

    Pad(8),
    #Reflection Properties
    reflection,
    Pad(16),

    Float("unknown0", VISIBLE=False),
    BytesRaw("unknown1", SIZE=16, VISIBLE=False),  # little endian dependency

    SIZE=400
    )


soso_body = Struct("tagdata",
    shdr_attrs,
    soso_attrs
    )


def get():
    return soso_def

soso_def = TagDef("soso",
    blam_header('soso', 2),
    soso_body,

    ext=".shader_model", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

messaging_information = Struct("messaging_information",
    SInt16("sequence_index"),
    SInt16("width_offset"),
    QStruct("offset_from_reference_corner",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    #QStruct("override_icon_color", INCLUDE=argb_byte),
    UInt32("override_icon_color", INCLUDE=argb_uint32),
    SInt8("frame_rate", MIN=0, MAX=30, UNIT_SCALE=per_sec_unit_scale),
    Bool8("flags",
        "use_text_from_string_list_instead",
        "override_default_color",
        "width_offset_is_absolute_icon_width",
        ),
    SInt16("text_index"),
    SIZE=64
    )

effector = Struct("effector",
    Pad(64),
    SEnum16("destination_type",
        "tint_0_to_1",
        "horizontal_offset",
        "vertical_offset",
        "fade_0_to_1",
        ),
    SEnum16("destination",
        "geometry_offset",
        "primary_map",
        "secondary_map",
        "tertiary_map",
        ),
    SEnum16("source",
        "player_pitch",
        "player_tangent",
        "player_yaw",
        "weapon_total_ammo",
        "weapon_loaded_ammo",
        "weapon_heat",
        "explicit",  # use low bound
        "weapon_zoom_level",
        ),

    Pad(2),
    QStruct("in_bounds", INCLUDE=from_to,
        SIDETIP="source_units"),  # source units
    QStruct("out_bounds", INCLUDE=from_to, SIDETIP="pixels"),  # pixels

    Pad(64),
    QStruct("tint_color_lower_bound", INCLUDE=rgb_float),
    QStruct("tint_color_upper_bound", INCLUDE=rgb_float),
    Struct("periodic_functions", INCLUDE=anim_func_per_pha),
    SIZE=220
    )

multitex_overlay = Struct("multitex_overlay",
    Pad(2),
    SInt16("type"),
    SEnum16("framebuffer_blend_func", *framebuffer_blend_functions),

    # Anchors
    Pad(34),
    SEnum16("primary_anchor", *multitex_anchors),
    SEnum16("secondary_anchor", *multitex_anchors),
    SEnum16("tertiary_anchor", *multitex_anchors),
    # Blending function
    SEnum16("zero_to_one_blend_func", *blending_funcs),
    SEnum16("one_to_two_blend_func", *blending_funcs),

    # Map scales
    Pad(2),
    QStruct("primary_scale", INCLUDE=xy_float),
    QStruct("secondary_scale", INCLUDE=xy_float),
    QStruct("tertiary_scale", INCLUDE=xy_float),
    # Map offsets
    QStruct("primary_offset", INCLUDE=xy_float),
    QStruct("secondary_offset", INCLUDE=xy_float),
    QStruct("tertiary_offset", INCLUDE=xy_float),

    # Maps
    dependency("primary_map", "bitm"),
    dependency("secondary_map", "bitm"),
    dependency("tertiary_map", "bitm"),
    SEnum16("primary_wrap_mode", *multitex_wrap_modes),
    SEnum16("secondary_wrap_mode", *multitex_wrap_modes),
    SEnum16("tertiary_wrap_mode", *multitex_wrap_modes),

    Pad(186),
    reflexive("effectors", effector, 30),
    SIZE=480,
    )

total_grenades_numbers = Struct("total_grenades_numbers",
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),

    Pad(4),
    SInt8("maximum_number_of_digits"),
    Bool8("flags",
        "show_leading_zeros",
        "only_show_when_zoomed",
        "draw_a_trailing_m",
        ),
    SInt8("number_of_fractional_digits"),

    Pad(13),
    SInt16("flash_cutoff"),
    SIZE=88
    )

overlay = Struct("overlay",
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),

    Pad(4),
    float_sec("frame_rate"),
    SInt16("sequence_index"),
    Bool16("type",
        "show_on_flashing",
        "show_on_empty",
        "show_on_default",
        "show_always",
        ),
    Bool32("flags",
        "flashes_when_active",
        ),

    SIZE=136
    )

warning_sound = Struct("warning_sound",
    dependency("sound", ('lsnd', 'snd!')),
    Bool32("latched_to",
        "low_grenade_sound",
        "no_grenades_left",
        "throw_on_no_grenades",
        ),
    Float("scale"),
    SIZE=56
    )

# Use this with INCLUDE keywords since it will need to be named
hud_background = Struct("",
    QStruct("anchor_offset",
        SInt16("x"), SInt16("y"), ORIENT='h'
        ),
    Float("width_scale"),
    Float("height_scale"),
    Bool16("scaling_flags", *hud_scaling_flags),

    Pad(22),
    dependency("interface_bitmap", "bitm"),
    #QStruct("default_color", INCLUDE=argb_byte),
    #QStruct("flashing_color", INCLUDE=argb_byte),
    UInt32("default_color", INCLUDE=argb_uint32),
    UInt32("flashing_color", INCLUDE=argb_uint32),
    float_sec("flash_period"),
    float_sec("flash_delay"),
    SInt16("number_of_flashes"),
    Bool16("flash_flags", *hud_flash_flags),
    float_sec("flash_length"),
    #QStruct("disabled_color", INCLUDE=argb_byte),
    UInt32("disabled_color", INCLUDE=argb_uint32),

    Pad(4),
    SInt16("sequence_index"),

    Pad(2),
    reflexive("multitex_overlays", multitex_overlay, 30),
    Pad(4),
    SIZE=104
    )

grhi_body = Struct("tagdata",
    SEnum16("anchor", *hud_anchors),

    Pad(34),
    Struct("grenade_hud_background", INCLUDE=hud_background),
    Struct("total_grenades",
        Struct("background", INCLUDE=hud_background),
        Struct("numbers", INCLUDE=total_grenades_numbers),
        dependency("overlay_bitmap", "bitm"),
        reflexive("overlays", overlay, 16),
        ),
    reflexive("warning_sounds", warning_sound, 12,
        DYN_NAME_PATH='.sound.filepath'),

    Pad(68),
    messaging_information,
    SIZE=504,
    )


def get():
    return grhi_def

grhi_def = TagDef("grhi",
    blam_header("grhi"),
    grhi_body,

    ext=".grenade_hud_interface", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .obje import *
from .objs.obje import ObjeTag

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=5)

responses = (
    "disappear",
    "detonate",
    "reflect",
    "overpenetrate",
    "attach"
    )

material_response = Struct("material_response",
    Bool16("flags",
        "cannot_be_overpenetrated",
        ),
    SEnum16('response', *responses),
    dependency('effect', "effe"),

    Pad(16),
    Struct("potential_response",
        SEnum16('response', *responses),
        Bool16("flags",
            "only_against_units",
            ),
        float_zero_to_one("skip_fraction"),
        from_to_rad("impact_angle"),  # radians
        from_to_wu_sec("impact_velocity"),  # world units/second
        dependency('effect', "effe"),
        ),

    Pad(16),
    SEnum16("scale_effects_by",
        "damage",
        "angle",
        ),
    Pad(2),
    float_rad("angular_noise"),
    float_wu_sec("velocity_noise"),
    dependency('detonation_effect', "effe"),

    Pad(24),
    # Penetration
    Float("initial_friction", UNIT_SCALE=per_sec_unit_scale),
    Float("maximum_distance"),

    # Reflection
    Float("parallel_refriction",    UNIT_SCALE=per_sec_unit_scale),
    Float("perpendicular_friction", UNIT_SCALE=per_sec_unit_scale),

    SIZE=160
    )

proj_attrs = Struct("proj_attrs",
    Bool32("flags",
        "oriented_along_velocity",
        "ai_must_use_ballistic_aiming",
        "detonation_max_time_if_attached",
        "has_super_combining_explosion",
        "add_parent_velocity_to_initial_velocity",
        "random_attached_detonation_time",
        "minimum_unattached_detonation_time",
        ),
    SEnum16('detonation_timer_starts',
        "immediately",
        "on_first_bounce",
        "when_at_rest",
        ),
    SEnum16('impact_noise', *sound_volumes),
    SEnum16('A_in', *projectile_inputs),
    SEnum16('B_in', *projectile_inputs),
    SEnum16('C_in', *projectile_inputs),
    SEnum16('D_in', *projectile_inputs),
    dependency('super_detonation', "effe"),
    float_wu("ai_perception_radius"),
    float_wu("collision_radius"),

    Struct("detonation",
        float_sec("arming_time"),
        float_wu("danger_radius"),
        dependency('effect', "effe"),
        from_to_sec("timer"),
        float_wu_sec("minimum_velocity"),
        float_wu("maximum_range"),
        ),

    Struct("physics",
        Float("air_gravity_scale"),
        from_to_wu("air_damage_range"),
        Float("water_gravity_scale"),
        from_to_wu("water_damage_range"),
        float_wu_sec("initial_velocity"),  # world units/sec
        float_wu_sec("final_velocity"),  # world units/sec
        float_rad_sec("guided_angular_velocity"),  # radians/second
        SEnum16('detonation_noise', *sound_volumes),

        Pad(2),
        dependency('detonation_started', "effe"),
        dependency('flyby_sound', "snd!"),
        dependency('attached_detonation_damage', "jpt!"),
        dependency('impact_damage', "jpt!"),
        ),

    Pad(12),
    reflexive("material_responses", material_response,
        len(materials_list), *materials_list),

    SIZE=208
    )

proj_body = Struct("tagdata",
    obje_attrs,
    proj_attrs,
    SIZE=588,
    )


def get():
    return proj_def

proj_def = TagDef("proj",
    blam_header('proj', 5),
    proj_body,

    ext=".projectile", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Veegie
# 	Completed plugin
# revision: 3		author: Lord Zedd
# 	Updated and standardized and restored shit veegie deleted
# revision: 4		author: Lord Zedd
# 	Now with comments.
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

jpt__category = (
    "none",
    "falling",
    "bullet",
    "grenade",
    "high_explosive",
    "sniper",
    "melee",
    "flame",
    "mounted_weapon",
    "vehicle",
    "plasma",
    "needle",
    "shotgun",
    )

jpt__player_response_priority = (
    "low",
    "medium",
    "high",
    )

jpt__player_response_type = (
    "none",
    "lighten",
    "darken",
    "max",
    "min",
    "invert",
    "tint",
    )

jpt__side_effect = (
    "none",
    "harmless",
    "lethal_to_the_unsuspecting",
    "emp",
    )


jpt__player_response = Struct("player_response",
    SEnum16("response_type", *jpt__player_response_response_type),
    SInt16("unknown_0", VISIBLE=False),
    SEnum16("type", *jpt__player_response_type),
    SEnum16("priority", *jpt__player_response_priority),
    Float("duration_0"),
    SEnum16("fade_function", *jpt__player_response_fade_function),
    SInt16("unknown_1", VISIBLE=False),
    Float("maximum_intensity"),
    color_argb_float("color"),
    Float("low_frequency_vibration_duration"),
    h3_rawdata_ref("low_frequency_vibration_function"),
    Float("high_frequency_vibration_duration"),
    h3_rawdata_ref("high_frequency_vibration_function"),
    h3_string_id("effect_name"),
    Float("duration_1"),
    h3_rawdata_ref("effect_scale_function"),
    ENDIAN=">", SIZE=112
    )


jpt__body = Struct("tagdata",
    QStruct("radius", INCLUDE=from_to),
    Float("cutoff_scale"),
    Bool32("flags_0",
        "dont_scale_damage_by_distance",
        "area_damage_players_only",
        ),
    SEnum16("side_effect", *jpt__side_effect),
    SEnum16("category", *jpt__category),
    Bool32("flags_1",
        "does_not_hurt_owner",
        "can_cause_headshots",
        "pings_resistant_units",
        "does_not_hurt_friends",
        "does_not_ping_units",
        "detonates_explosives",
        "only_hurts_shields",
        "causes_flaming_death",
        "damage_indicators_always_point_down",
        "skips_shields",
        "only_hurts_one_infection_form",
        ("infection_form_pop", 1 << 12),
        "ignore_seat_scale_for_direct_damage",
        "forces_hard_ping",
        "does_not_hurt_players",
        ),
    Float("area_of_effect_core_radius"),
    Float("damage_lower_bound"),
    QStruct("damage_upper_bound", INCLUDE=from_to),
    float_rad("damage_inner_cone_angle"),
    float_rad("damage_outer_cone_angle"),
    Float("active_camoflage_damage"),
    Float("stun"),
    Float("max_stun"),
    Float("stun_time"),
    Float("instantaneous_acceleration"),
    Float("rider_direct_damage_scale"),
    Float("rider_max_transfer_damage_scale"),
    Float("rider_min_transfer_damage_scale"),
    h3_string_id("general_damage"),
    h3_string_id("specific_damage"),
    h3_string_id("special_damage"),
    Float("ai_stun_radius"),
    QStruct("ai_stun_bounds", INCLUDE=from_to),
    Float("shake_radius"),
    Float("emp_radius"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    h3_reflexive("player_responses", jpt__player_response),
    h3_dependency("damage_response"),
    Float("duration_0"),
    SEnum16("fade_function", *jpt__player_response_fade_function),
    SInt16("unknown_2", VISIBLE=False),
    float_rad("rotation"),
    Float("pushback"),
    QStruct("jitter", INCLUDE=from_to),
    Float("duration_1"),
    SEnum16("falloff_function", *jpt__player_response_fade_function),
    SInt16("unknown_3", VISIBLE=False),
    Float("random_translation"),
    float_rad("random_rotation"),
    SEnum16("wobble_function", *jpt__wobble_function),
    SInt16("unknown_4", VISIBLE=False),
    Float("wobble_function_period"),
    Float("wobble_weight"),
    h3_dependency("sound"),
    Float("forward_velocity"),
    Float("forward_radius"),
    Float("forward_exponent"),
    Float("outward_velocity"),
    Float("outward_radius"),
    Float("outward_exponent"),
    ENDIAN=">", SIZE=240
    )


def get():
    return jpt__def

jpt__def = TagDef("jpt!",
    h3_blam_header('jpt!'),
    jpt__body,

    ext=".%s" % h3_tag_class_fcc_to_ext["jpt!"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


trak_camera_point = Struct("camera_point",
    QStruct("position", INCLUDE=ijk_float, VISIBLE=False),
    QStruct("orientation", INCLUDE=ijkw_float, VISIBLE=False),
    ENDIAN=">", SIZE=28
    )


trak_body = Struct("tagdata",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    h3_reflexive("camera_points", trak_camera_point),
    ENDIAN=">", SIZE=16
    )


def get():
    return trak_def

trak_def = TagDef("trak",
    h3_blam_header('trak'),
    trak_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["trak"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: -DeToX-
# 	Created layout of plugin
# revision: 2		author: Lord Zedd
# 	Done
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


cddf_body = Struct("tagdata",
    Float("apply_damage_scale"),
    Float("apply_recoil_damage_scale"),
    QStruct("damage_acceleration", INCLUDE=from_to),
    Float("damage_scale_min_0"),
    Float("damage_scale_min_1"),
    QStruct("recoil_damage_acceleration", INCLUDE=from_to),
    QStruct("recoil_damage_scale", INCLUDE=from_to),
    ENDIAN=">", SIZE=40
    )


def get():
    return cddf_def

cddf_def = TagDef("cddf",
    h3_blam_header('cddf'),
    cddf_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["cddf"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from collections import OrderedDict
from traceback import format_exc

from array import array
from reclaimer.common_descs import rawdata_ref_struct
from reclaimer.field_types import FieldType, RawdataRef, Reflexive, TagRef
from supyr_struct.defs.block_def import BlockDef
from supyr_struct.tag import Tag
from ..wrappers.map_pointer_converter import MapPointerConverter

empty_rawdata_def = BlockDef("samples_stub", INCLUDE=rawdata_ref_struct)

REMOVED_RSRC_TAG_DIR = "!rem\\"


def set_tag_meta_pointers(block, meta_pointer, raw_pointer_base=0):
    if not (hasattr(block, "desc") and hasattr(block, "__iter__")):
        return meta_pointer, raw_pointer_base

    item_indices = list(range(len(block)))
    if hasattr(block, "STEPTREE"):
        item_indices.append("STEPTREE")

    for i in item_indices:
        typ = block.get_desc("TYPE", i)
        sub_block = block[i]
        if typ == RawdataRef:
            if getattr(sub_block, "data", None):
                orig_size = sub_block.size
                # pad data streams to multiples of 4 bytes for alignment
                sub_block.data += b'\x00' * ((4 - (sub_block.size % 4)) % 4)
                sub_block.size = orig_size

            sub_block.id = sub_block.pointer = 0
            if not sub_block.flags.data_in_resource_map:
                sub_block.pointer = meta_pointer
                sub_block.raw_pointer = meta_pointer + raw_pointer_base
                meta_pointer += len(getattr(sub_block, "data", ()))
            else:
                raw_pointer_base += sub_block.size

        elif typ == TagRef:
            sub_block.filepath = ""
            sub_block.id = 0xFFffFFff
        elif typ == Reflexive:
            sub_block.pointer = meta_pointer
            sub_block.id = 0
            if sub_block.STEPTREE:
                meta_pointer += sub_block.size * sub_block.STEPTREE[0].SIZE

        meta_pointer, raw_pointer_base = set_tag_meta_pointers(
            sub_block, meta_pointer, raw_pointer_base)

    meta_pointer += (4 - (meta_pointer % 4)) % 4
    return meta_pointer, raw_pointer_base


class Halo1RsrcMapTag(Tag):
    defs = ()
    tag_path_indices = ()

    _orig_tag_count = 0

    def __init__(self, *args, **kwargs):
        Tag.__init__(self, *args, **kwargs)
        self.map_pointer_converter = MapPointerConverter((0, 0, 0xFFffFFff))
        self.setup_defs()

        try:
            rsrc_tags = self.data.tags
        except Exception:
            rsrc_tags = ()
        self.tag_path_indices = {rsrc_tags[i].tag.path.lower(): i
                                 for i in range(len(rsrc_tags))}

    def parse(self, *args, **kwargs):
        Tag.parse(self, *args, **kwargs)
        try:
            self._orig_tag_count = len(self.data.tags)
        except Exception:
            self._orig_tag_count = 0

    @property
    def orig_tag_count(self):
        return self._orig_tag_count

    def setup_defs(self):
        if Halo1RsrcMapTag.defs:
            return

        Halo1RsrcMapTag.defs = defs = {}
        for fcc in ("bitm", "snd!", "font", "hmt ", "ustr"):
            try:
                fcc2 = fcc.replace("!", "_").replace(" ", "_")
                exec("from reclaimer.hek.defs.%s import %s_def" %
                     (fcc2, fcc2))
                exec("defs['%s'] = %s_def" % (fcc, fcc2))
            except Exception:
                print(format_exc())

    def load_resource_tags(self, *tags_dirs, ignored=(),
                           tag_paths_to_load_by_tags_dirs=()):
        tag_paths_to_load_by_tags_dirs = dict(tag_paths_to_load_by_tags_dirs)
        loaded_tags = OrderedDict()
        usable_defs = {}
        rsrc_type = self.data.resource_type.enum_name
        if rsrc_type == "bitmaps":
            usable_defs["bitmap"] = self.defs.get("bitm")
        elif rsrc_type == "sounds":
            usable_defs["sound"] = self.defs.get("snd!")
        elif rsrc_type == "strings":
            usable_defs["font"] = self.defs.get("font")
            usable_defs["hud_message_text"] = self.defs.get("hmt ")
            usable_defs["unicode_string_list"] = self.defs.get("ustr")

        for tags_dir in tags_dirs:
            tags_to_load = []
            tag_paths_to_load_by_tags_dirs[tags_dir] = tags_to_load
            for root, dirs, files in os.walk(tags_dir):
                for filename in sorted(files):
                    filepath = os.path.join(root, filename)
                    tag_path, ext = os.path.splitext(
                        filepath.lower().split(tags_dir.lower())[-1])
                    ext = ext.strip(".")

                    if usable_defs.get(ext) and (tag_path not in ignored and
                                                 tag_path not in loaded_tags):
                        tags_to_load.append(tag_path + "." + ext)

        for tags_dir, tag_paths in tag_paths_to_load_by_tags_dirs.items():
            for tag_path in tag_paths:
                try:
                    filepath = os.path.join(tags_dir, tag_path)
                    tag_path, ext = os.path.splitext(
                        filepath.lower().split(tags_dir.lower())[-1])
                    ext = ext.strip(".")
                    loaded_tags[tag_path] = usable_defs[ext].build(
                        filepath=filepath)
                except Exception:
                    print(format_exc())
                    print("Could not load: '%s'" % tag_path)

        return loaded_tags

    def set_pointers(self, offset=0):  # offset is unused
        resource_tags = self.data.tags
        curr_data_pointer = 16
        curr_path_pointer = 0
        for i in range(len(resource_tags)):
            tag_head = resource_tags[i]
            tag_head.path_offset = curr_path_pointer
            curr_path_pointer += len(tag_head.tag.path) + 1

            # align the data pointer
            curr_data_pointer += ((4 - (curr_data_pointer % 4)) % 4)
            if i >= self.orig_tag_count:
                # only adjust the data offset of new tags
                tag_head.offset = curr_data_pointer

            curr_data_pointer = tag_head.offset + tag_head.size

        curr_data_pointer += (4 - (curr_data_pointer % 4)) % 4
        self.data.tag_paths_pointer   = curr_data_pointer
        self.data.tag_headers_pointer = (curr_data_pointer + curr_path_pointer)

    def update_bitmap(self, tag_path, new_tag):
        if self.data.resource_type.enum_name != "bitmaps":
            raise TypeError("Cannot add bitmap to %s resource map" %
                            self.data.resource_type.enum_name)

        meta_head_i = self.tag_path_indices.get(tag_path)
        pixel_head_i = self.tag_path_indices.get(tag_path + "__pixels")
        if None in (meta_head_i, pixel_head_i):
            raise ValueError("'%s' is not in resource map" % tag_path)
        elif self.defs.get("bitm", None) is None:
            raise TypeError("bitmap tag definition is not loaded.")

        pix_head = self.data.tags[pixel_head_i]
        new_bitmaps = new_tag.data.tagdata.bitmaps.STEPTREE
        new_pixels = new_tag.data.tagdata.processed_pixel_data.data

        desc  = self.defs['bitm'].descriptor[1]
        parent = [None]
        with FieldType.force_little:
            desc['TYPE'].parser(
                desc, parent=parent, attr_index=0, parsing_resource=True,
                rawdata=self.data.tags[meta_head_i].tag.data,
                map_pointer_converter=self.map_pointer_converter)
            orig_bitmaps = parent[0].bitmaps.STEPTREE

        if len(orig_bitmaps) != len(new_bitmaps):
            raise ValueError(
                "%s:\nReplacement bitmap count(%s) does not match original(%s)"
                % (tag_path, len(new_bitmaps), len(orig_bitmaps)))

        for i in range(len(new_bitmaps)):
            new_bitmap = new_bitmaps[i]
            orig_bitmap = orig_bitmaps[i]
            for attr_name in ("width", "height", "depth", "type", "format"):
                new_bitm_attr = new_bitmap[attr_name]
                orig_bitm_attr = orig_bitmap[attr_name]
                if attr_name in ("type", "format"):
                    new_bitm_attr = new_bitm_attr.enum_name
                    orig_bitm_attr = orig_bitm_attr.enum_name

                if new_bitm_attr != orig_bitm_attr:
                    raise ValueError(
                        "%s:\nReplacement bitmap %s(%s) does not match original(%s)"
                        % (tag_path, attr_name, new_bitm_attr, orig_bitm_attr))

        if len(new_bitmaps) == 1 and len(pix_head.tag.data) > len(new_pixels):
            # super hacky, but it'll work for now. copy the end of the mipmap
            # data from the old bitmap to the end of the new pixel data
            new_pixels += pix_head.tag.data[len(new_pixels): ]

        if len(pix_head.tag.data) != len(new_pixels):
            raise ValueError(
                "%s:\nReplacement data length(%s) does not match original(%s)"
                % (tag_path, len(new_pixels), len(pix_head.tag.data)))
        else:
            pix_head.tag.data = new_pixels

    def _add_bitmap(self, tag_path, new_tag, base_pointer, depreciate):
        tag_data = new_tag.data.tagdata
        rsrc_tags = self.data.tags
        src_pixels = tag_data.processed_pixel_data.data

        # add headers for this tag
        curr_idx = len(rsrc_tags)
        rsrc_tags.extend(2)
        meta_head, pix_head = rsrc_tags[-1], rsrc_tags[-2]

        for i in range(len(tag_data.sequences.STEPTREE)):
            seq = tag_data.sequences.STEPTREE[-1]
            if seq.sprites.STEPTREE:
                seq.first_bitmap_index = 0
                seq.bitmap_count = 0
            elif not (seq.first_bitmap_index in range(tag_data.bitmaps.size)):
                tag_data.sequences.STEPTREE.pop(-1)
                continue

            break

        # fill out each bitmap block's flags, unknowns, and sizes
        for bitmap in tag_data.bitmaps.STEPTREE:
            min_dim = 4 if "dxt" in bitmap.format.enum_name else 1
            w, h, d = bitmap.width, bitmap.height, bitmap.depth
            pix_chunk_size = 0
            bpp = 8
            if bitmap.format.data in (3, 6, 8, 9):
                bpp = 16
            elif bitmap.format.data in (10, 11):
                bpp = 32
            elif bitmap.format.data == 14:
                bpp = 4

            for i in range(bitmap.mipmaps + 1):
                pix_chunk_size += (w * h * d * bpp) // 8
                w = max(w >> 1, 1)
                h = max(h >> 1, 1)
                d = max(d >> 1, 1)
                w += (min_dim - (w % min_dim)) % min_dim
                h += (min_dim - (h % min_dim)) % min_dim

            if bitmap.type.enum_name == "cubemap":
                pix_chunk_size *= 6

            bitmap.flags.data_in_resource_map = True
            bitmap.flags.prefer_low_detail = True

            bitmap.pixels_meta_size = pix_chunk_size
            bitmap.bitmap_data_pointer = 0xFFffFFff
            bitmap.bitmap_id_unknown1 = 0  # don't know how to set this
            bitmap.bitmap_id_unknown2 = bitmap.base_address = 0

        # fill out each bitmap block pixel data pointer and copy the pixels
        # from each bitmap out of the src_pixels and piece them together
        pixel_off = 0
        combined_pixels = b''
        for bitmap in tag_data.bitmaps.STEPTREE:
            combined_pixels += src_pixels[
                bitmap.pixels_offset:
                bitmap.pixels_offset + bitmap.pixels_meta_size]

            bitmap.pixels_offset = base_pointer + pixel_off
            pixel_off += bitmap.pixels_meta_size

        if tag_path in self.tag_path_indices:
            old_idx = self.tag_path_indices[tag_path]
            old_meta_head, old_pix_head = rsrc_tags[old_idx], rsrc_tags[old_idx - 1]
            if not depreciate or old_pix_head.tag.data == combined_pixels:
                # same pixel data. optimize away
                del rsrc_tags[-2: ]
                return None, None

            old_pix_head.tag.path = REMOVED_RSRC_TAG_DIR + tag_path + "__pixels"
            old_meta_head.tag.path = REMOVED_RSRC_TAG_DIR + tag_path

        # null these to prevent them from screwing with the size calculations
        tag_data.compressed_color_plate_data.data = b""
        tag_data.processed_pixel_data.data = b""

        pix_head.tag.data = combined_pixels
        return pix_head, meta_head

    def _add_sound(self, tag_path, new_tag, base_pointer, depreciate):
        tag_data = new_tag.data.tagdata
        rsrc_tags = self.data.tags
        if tag_data.promotion_sound.filepath:
            # not gonna bother adding sounds with dependencies
            return None, None
        elif tag_path in self.tag_path_indices:
            return None, None

        max_bend = tag_data.maximum_bend_per_second
        if max_bend != 0:
            tag_data.maximum_bend_per_second = 1 / (max_bend ** 30)

        samples = b''
        # add headers for this tag
        rsrc_tags.extend(2)
        meta_head, samp_head = rsrc_tags[-1], rsrc_tags[-2]

        for pr in tag_data.pitch_ranges.STEPTREE:
            pr.unknown0 = 1.0
            pr.unknown1 = pr.unknown2 = -1
            for perm in pr.permutations.STEPTREE:
                perm.unknown1 = 0
                perm.unknown3 = perm.unknown2 = 0xFFffFFff
                sample_data = perm.samples.data
                if perm.compression.enum_name == "none":
                    sample_data = array(">h", sample_data)
                    sample_data.byteswap()
                    sample_data = sample_data.tobytes()

                perm.samples.flags.data_in_resource_map = True
                perm.mouth_data.flags.data_in_resource_map = False
                perm.subtitle_data.flags.data_in_resource_map = False

                perm.samples.raw_pointer = base_pointer
                base_pointer += len(sample_data)

                samples += sample_data
                perm.samples = empty_rawdata_def.build(initdata=perm.samples)

        samp_head.tag.data = samples

        return samp_head, meta_head

    def add_tag(self, tag_path, new_tag, base_pointer=None, depreciate=False):
        tag_data = new_tag.data.tagdata
        if base_pointer is None:
            base_pointer = self.data.tag_paths_pointer

        # make sure the base pointer is aligned
        base_pointer += ((4 - (base_pointer % 4)) % 4)

        rsrc_type = self.data.resource_type.enum_name
        raw_head = meta_head = None

        if rsrc_type == "bitmaps":
            raw_head, meta_head = self._add_bitmap(tag_path, new_tag,
                                                   base_pointer, depreciate)
            if not (raw_head and meta_head):
                return base_pointer
            raw_head.tag.path = tag_path + "__pixels"
            set_tag_meta_pointers(tag_data, tag_data.SIZE)

        elif rsrc_type == "sounds":
            if tag_data.promotion_sound.filepath:
                # not gonna bother adding sounds with dependencies
                raise TypeError("Cannot add sound tag with promotion sound.")

            raw_head, meta_head = self._add_sound(tag_path, new_tag,
                                                  base_pointer, depreciate)
            if not (raw_head and meta_head):
                return base_pointer
            raw_head.tag.path = tag_path + "__permutations"

            set_tag_meta_pointers(tag_data, 0, tag_data.SIZE)
        else:
            set_tag_meta_pointers(tag_data, tag_data.SIZE)

        # fill out the tag headers
        meta_head.tag.path = tag_path
        with FieldType.force_little:
            meta_head.tag.data = tag_data.serialize()

        base_pointer += meta_head.size
        base_pointer += ((4 - (base_pointer % 4)) % 4)
        if raw_head:
            base_pointer += raw_head.size

        return base_pointer + ((4 - (base_pointer % 4)) % 4)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# ##############
#   metadata   #
# ##############
__author__ = "Devin Bobadilla, Michelle van der Graaf"
#           YYYY.MM.DD
__date__ = "2020.02.14"
__version__ = (2, 7, 2)
__website__ = "https://github.com/Sigmmma/reclaimer"
__all__ = (
    "animation", "bitmaps", "h2", "h3", "halo_script", "hek", "meta", "misc",
    "model", "os_hek", "os_v3_hek", "os_v4_hek", "physics",
    "shadowrun_prototype", "sounds", "strings", "stubbs", "util",

    "common_descs", "constants", "data_extraction", "enums",
    "field_type_methods", "field_types",
    )
